# -*- coding: utf-8 -*-
#
# This file is part of S4D.
#
# SD4 is a python package for speaker diarization based on SIDEKIT.
# S4D home page: http://www-lium.univ-lemans.fr/s4d/
# SIDEKIT home page: http://www-lium.univ-lemans.fr/sidekit/
#
# S4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# S4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

"""
Diar is a class describing an audio/video segmentation file. A diarization
contains a list of segments. Where each row is segment composed of n values
identified by a attribut names.
The diarization file is the most important file in the toolkit. All programs
are driven by a diarization file and most of them generate a diarization
file (trainer generate gmm).

A diarization stores a list of ''segments'' composed of attributes.

A diarization could draw data from several shows. It is very useful in a batch
mode context (training of GMM, computing log likelihood ratio, cross-show
diarization, etc.).

Example
-------
>>> diarization[0] //get the first segment
['20041006_0700_0800_CLASSIQUE', 'Emmanuel_Cugny', 'speaker', 164, 1170]
>>> diarization[0]['show']
'20041006_0700_0800_CLASSIQUE'
>>> diarization[0]['cluster']
'Emmanuel_Cugny'
>>> diarization[0]['start']
164
>>> diarization[0]['stop']
1170

where:
  * attribut 0 named ''show'': ''20041006_0700_0800_CLASSIQUE'' = the show speaker
  * attribut 1 named ''cluster'' : ''Emmanuel_Cugny'' the speaker speaker
  * attribut 2 named ''type'' : ''speaker'' contains the cluster type (speaker or head)
  * attribut 3 named ''start'': ''164'' is the index of the first feature of the segment
  * attribut 4 named ''stop'': ''1170'' is the index of the last feature of the segment

How to
------
* Read a diarization:
    ::

        from s4d.diarization import Diar
        diarization = Diar.read_seg('foo.seg') //LIUM Spk Diarization format
        diarization = Diar.read_mdtm('foo.seg') //MDTM format
        diarization = Diar.read_rttm('foo.seg') //RTTM format
        diarization = Diar.read_uem('foo.seg') //UEM format

* Get a segment:
    ::

        seg = diarization[0]

* Get a attribut of a segment
    ::

        seg['cluster']

* Write a diarization:
    ::

        diarization = Diar.write_seg('foo.seg', diarization) //LIUM format

* Add or remove an attribut:
    ::

        diarization.add_attribut(speaker='gender', default=None) // add an attribut named 'gender, the default value is None
        diarization.del_attribut('gender') // remove the attribut

* Sort a diarization:
    ::

        diarization.sort()

* Create a new segment:
    ::

        diarization.append(show='foo', cluster='speaker', start=0, stop=100)

Add all the segments of diar2 into diar1:
    ::

        diar1.append_diar(diar2)

Modules
-------

"""



import copy
import logging
import os
import re
import sys


from sidekit.sidekit_wrappers import *
from sidekit.bosaris.idmap import IdMap
from six import string_types
from .utils import str2str_normalize

try:
    from sortedcontainers import SortedDict as dict
except ImportError:
    pass


class Diar():
    """
    The diarization class.

    :attr _attributes: a AttributeNames object storing the attribut definitions
    :attr cluster_types: a list object
    :attr segments: a list of Segment object
    """
    def __init__(self):
        self._attributes = AttributeNames()
        self._attributes.initialize({'show': 0, 'cluster': 1, 'cluster_type': 2,
                                  'start': 3, 'stop': 4},
                                    ['empty', 'empty', 'speaker', 0, 0])
        self.cluster_types = ['speaker', 'head']
        self.segments = list()

    def copy_structure(self):
        """
        Copy the internal structure of the diarization, ie the attribut names
        and the cluster types. The data is not copy.
        :return: a Diar object
        """
        tmp_diarization = Diar()
        tmp_diarization._attributes = copy.deepcopy(self._attributes)
        tmp_diarization.cluster_types = copy.deepcopy(self.cluster_types)
        return tmp_diarization

    def del_all(self, attribute, value):
        """
        Delete all segments satisfing the boolean expression [attribute = value]
        :param attribute: speaker of the attribute to delete
        :param value:
        :return:
        """
        lst = list()
        for segment in self.segments:
            if segment[attribute] != value:
                lst.append(segment)
        self.segments = lst

    def overlap(self, add_intersection=False):
        """
        remove overlap zone
        :return: a new diarization without overlap
        """

        def add(_show, _features_index, _cluster_list, _uem):
            diar_tmp = self.copy_structure()
            for lcluster in _cluster_list:
                lst = sorted(_uem.intersection(set(_features_index[lcluster])))
                if len(lst) > 0:
                    c = lst[0];
                    diar_tmp.append(show=_show, start=c, stop=c+1, cluster=lcluster)
                    l = 0
                    for j in range(1, len(lst)):
                        p = c
                        c = lst[j]
                        if c == p + 1:
                            l += 1
                        else:
                            diar_tmp[-1]['stop'] += l
                            diar_tmp.append(show=_show, start=c, stop=c+1, cluster=lcluster)
                            l = 0
                    if l > 0:
                        diar_tmp[-1]['stop'] += l
            return diar_tmp

        diarization_out = self.copy_structure()
        #shows = self.unique('show')
        shows = self.make_index(['show'])
        for show in shows:
            logging.info('rm overlap show: '+show)
            diar_show = shows[show]
            length = diar_show.last_feature_index()
            cluster_list = diar_show.unique('cluster')
            mat = numpy.zeros(length)
            features_index = diar_show.features_by_cluster()
            for i, cluster in enumerate(cluster_list):
                mat[features_index[cluster]] += 1
            uem = set(numpy.where(mat == 1)[0].tolist())
            diarization_out.segments += add(show, features_index, cluster_list, uem)
            if add_intersection:
                uem = set(numpy.where(mat != 1)[0].tolist())
                diarization_out.segments += add(show, features_index, cluster_list, uem)

        return diarization_out

    def filter(self, attribute, operator, value):
        """
        build a new diarization whose segments satisfy the boolean expression
        [attribute operator value]
        :param attribute: a attribute speaker (str)
        :param operator: a comperator opertor (> < >= <= in == !=)
        :param value: the value (int, float, str, list...)
        :return: a Diar object
        """
        tmp_diarization = self.copy_structure()
        tmp_diarization.segments = list()
        if attribute == "length" or attribute == "duration":
            str = "seg.duration() {:s} {}".format(operator, value)
        elif isinstance(value, string_types):
            str = "seg['{:s}'] {:s} '{:s}'".format(attribute, operator, value)
        else:
            str = "seg['{:s}'] {:s} {} ".format(attribute, operator, value)
        # print(ch)
        logging.debug(str)
        for seg in self.segments:
            # print(ch, seg.length())
            if eval(str):
                tmp_diarization.segments.append(copy.deepcopy(seg))
        return tmp_diarization

    def rename(self, attribute, old_values, new_value):
        """
        Rename all values in list old_values into the new value new_value
        :param attribute: speaker of the attribute
        :param old_values:  list of old values
        :param new_value: new value

        """
        for segment in self.segments:
            if segment[attribute] in old_values or len(old_values) == 0 :
                segment[attribute] = new_value

    def _iofi(self, index, attributes, segment):
        """
        recursive fonction to add a segment into the n level keys dictionary

        :param index: dict object of level n
        :param attributes: list of attribut attributes
        :param segment: a segment
        :return: a dictornary of level n that contains sub diarization. Segments
        are not copy.
        """
        # removes and gets the last attribut speaker
        attribut = attributes.pop()
        # takes the values of this attribut speaker
        value = segment[attribut]
        # if there is no more attribut attributes
        if len(attributes) <= 0:
            if value in index:
                # add the segment to the list
                index[value].append_seg(segment)
            else:
                # create a list and add the segment
                ldiar = self.copy_structure()
                ldiar.append_seg(segment)
                index[value] = ldiar
            return index
        else:
            # recursion to the level n-1 until attributes is empty
            self._iofi(index[value], attributes, segment)

    def make_index(self, attributes):
        """
        Build a n level key dictionary (dictionary of dictionaries of
        dictionaries...) based on Index.
        Index is an implementation of perl's autovivification feature.
        The values contains a list of row.

        example :

            d = make_index(['show', 'gender', 'cluster'])

            print(d['show1']['M']['speaker'])

        :param attributes: a list of attribut _attributes corresponding to the key indexes
        :return: a dictionary of sub diarization. Segments are not copy.
        """
        index = Index()
        for segment in self.segments:
            self._iofi(index, attributes[::-1], segment)
        return index

    def unique(self, attibute):
        """
        :param attibute: the attibute of the attribut
        :return: a list object of unique value of the attribut
        """
        dic = dict()
        lst = list()
        for seg in self.segments:
            dic[seg[attibute]] = 0
        for value in dic.keys():
            lst.append(value)
        return lst

    def sort(self, attributes=['show', 'start'], reverse=False):
        """
        Sort the segments
        :param attributes: a list of attribut names
        :param reverse: if true, make a reverse sort

        """
        attributes.reverse()
        for attribute in attributes:
            if attribute not in self._attributes:
                raise Exception("This attribut don't exits : " + attribute)
            self.segments = sorted(self.segments, key=lambda x: x[self._attributes[attribute]],
                                   reverse=reverse)

    def clear(self):
        """
        remove all the segments
        :return:
        """
        self.segments = list()

    def add_attribut(self, new_attribut, default=''):
        """
        Add a attribut
        :param new_attribut: the speaker of the new attribut
        :param default: the default value of the attribut

        """
        self._attributes.add(new_attribut, default)
        for seg in self.segments:
            seg.append(default)

    def del_attribut(self, attribut):
        """
        Delete a attribut
        :param attribut: the speaker of the attribut to detele

        """
        if attribut not in self._attributes:
            raise Exception("This attribut don't exits : " + attribut)
        else:
            i = self._attributes[attribut]
            for seg in self.segments:
                del seg[i]
            self._attributes.delete(attribut)

    def _new_row(self, **kwargs):
        """
        Create a new segment initialized with kwargs
        :param kwargs: the values
        :return:
        """
        seg = Segment(self._attributes.defaults, self._attributes)
        for key, value in kwargs.items():
            seg[self._attributes[key]] = value
        return seg

    def append(self, **kwargs):
        """
        Transforme a list of values into a segment and append the segmnt into
        the existing segment list.
        :param kwargs: the values
        :return:
        """
        self.segments.append(self._new_row(**kwargs))

    def append_seg(self, segment):
        """
        Append a Segment object into the existing segment list.

        :param segment: a Segment object

        """
        self.segments.append(segment)

    def append_list(self, segment_lst):
        """
        Append a list of segments into the existing segment segment_lst.
        :param segment_lst: a list of segments

        """
        self.segments += segment_lst

    def append_diar(self, out_diarization):
        """
        Append a diarization.
        :param out_diarization: a diarization object

        """
        lMatchAttr=list()
        for i in self._attributes.names:
            if i in out_diarization._attributes.names:
                lMatchAttr.append(i)
        assert len(lMatchAttr)!=0,"No attribute matches"
        
        lSegments=list()
        for i in out_diarization:
            seg=Segment(self._attributes.defaults,self._attributes)
            for y in lMatchAttr:
                seg._set_attr(y,i[y])
            lSegments.append(seg)
        
        self.segments += lSegments

    def insert(self, i, **kwargs):
        """
        Insert values into the list at offset index
        :param i: This is the Index where the object obj need to be inserted.
        :param kwargs: the values

        """
        self.segments.insert(i, self._new_row(**kwargs))

    def __iter__(self):
        """
        This method is called when an iterator is required for a container.
        :return: an iterator
        """
        return self.segments.__iter__()

    def __reversed__(self):
        """
        Called (if present) by the reversed() built-in to implement reverse iteration.
        :return: a Diar object
        """
        return self.segments.__reversed__()

    def __delitem__(self, index):
        """
        Called to implement deletion of self[index]
        :param index: a int

        """
        del self.segments[index]

    def __getitem__(self, index):
        """
        Called to implement evaluation of self[index]
        :param index: a int
        :return: a Segment object
        """
        return self.segments[index]

    def __setitem__(self, index, value):
        """
        Called to implement evaluation of self[index] = value

        :param index: a int
        :param value:  a Segment object

        """
        self.segments[index] = value

    def __len__(self):
        """
        :return: the number of segments
        """
        return len(self.segments)

    def __eq__(self, diarization): # real signature unknown
        idx_self = self.make_index(['show', 'cluster', 'start'])
        idx = diarization.make_index(['show', 'cluster', 'start'])

        for show in idx_self:
            for cluster in idx_self[show]:
                for start in idx_self[show][cluster]:
                    if show not in idx:
                        return False
                    elif cluster not in idx[show]:
                        return False
                    elif start not in idx[show][cluster]:
                        return False
                    else :
                        l1 = idx[show][cluster][start].segments
                        l2 = idx_self[show][cluster][start].segments
                        return all(x in l2 for x in l1)
        return True

    def __ne__(self, diarization): # real signature unknown
        return not self.__eq__(diarization)

    def __repr__(self):
        """
        :return: a string version of the diarization
        """
        string = '  attribut definition  : ['
        index = 0
        lst = self._attributes.sorted()
        #print(lst)
        for attribute in lst:
            string += "'" + attribute[0] + "', "
        string = re.sub(', $', '', string) + ']\n'
        for segment in self.segments:
            line = ''
            for attribute in segment:
                line += attribute.__repr__() + ', '
            string += '  row ' + str(index) + ': [' + re.sub(', $', '',
                                                           line) + ']\n'
            index += 1
        return '[\n' + string + ']'

    def __add__(self, diarization):
        diarization_copy = copy.deepcopy(self)
        diarization_copy.segments += diarization.segments
        return diarization_copy

    def __iadd__(self, diarization):
        self.segments += diarization.segments
        return self

    def id_map(self, id_attribut='cluster', show_attribut='show',
               prefix_id_attrubut=None, suffix_show_attribut=None):
        """
        Generate a IdMap object for the StatServer
        :param id_attribut: speaker id_attribut attribut
        :param show_attribut: show_attribut attribut
        :param prefix_id_attrubut: prefix string of id_attribut
        :param suffix_show_attribut: suffix string of id_attribut

        :param out_diarization: a diarization object
        :return: a IdMap object
        """
        id_map = IdMap()
        id_map.leftids = numpy.empty(len(self.segments), dtype="|O")
        id_map.rightids = numpy.empty(len(self.segments), dtype="|O")
        id_map.start = numpy.empty(len(self.segments), dtype="|O")
        id_map.stop = numpy.empty(len(self.segments), dtype="|O")

        i = 0
        for segment in self.segments:
            if prefix_id_attrubut is not None:
                id_map.leftids[i] = segment[prefix_id_attrubut] + '/' + segment[id_attribut]
            else:
                id_map.leftids[i] = segment[id_attribut]
            if suffix_show_attribut is not None:
                id_map.rightids[i] = segment[show_attribut] + '/' + segment[suffix_show_attribut]
            else:
                id_map.rightids[i] = segment[show_attribut]
            id_map.start[i] = segment['start']
            id_map.stop[i] = segment['stop']
            i += 1

        return id_map

    def features_by_cluster(self, show=None, maximum_length=None):
        """
        Generate the indexes of a show
        :param show: the speaker of the show
        :param maximum_length: maximum length of the show
        :return: a dict object (keys are the cluster_list)
        """
        if show == None:
            l = self.unique('show')
            if len(l) > 1:
                raise Exception('diarization address sevreal shows, set show parameter')
            else:
                show = l[0]
        dic = dict()
        for segment in self.segments:
            if show == segment['show']:
                cluster = segment['cluster']
                start = segment['start']
                stop = segment['stop']
                if maximum_length is not None:
                    start = min(segment['start'], maximum_length)
                    stop = min(segment['stop'], maximum_length)

                if cluster not in dic:
                    dic[cluster] = []
                dic[cluster] += [i for i in range(start, stop)]
        return dic

    def features(self, show=None, maximum_length=None):
        """
        Generate the index features of a show
        :param show: a string corresponding to the speaker of the show
        :param maximum_length: maximum length of the show
        :return: a list object of indexes
        """
        if show is None:
            lst = self.unique('show')
            if len(lst) > 1:
                raise Exception('diarization address sevreal shows, set show parameter')
            else:
                show = lst[0]
        lst = list()
        for segment in self.segments:
            if show == segment['show']:
                start = segment['start']
                stop = segment['stop']
                if maximum_length is not None:
                    start = min(segment['start'], maximum_length)
                    stop = min(segment['stop'], maximum_length)
                lst += [i for i in range(start, stop)]
        return lst
    def to_list(self, show=None, uem_start=None, uem_stop=None):
        if show is None:
            lst = self.unique('show')
            if len(lst) > 1:
                raise Exception('diarization address sevreal shows, set show parameter')
            else:
                show = lst[0]
        if uem_start is None:
            uem_start = 0
        if uem_stop is None:
            uem_stop = self.last_feature_index()
        lst = [''] * uem_stop
        for segment in self.segments:
            if show == segment['show']:
                cluster = segment['cluster']
                start = max(segment['start'], uem_start)
                stop = min(segment['stop'], uem_stop)
                for i in range(start, stop):
                    if lst[i] == '':
                        lst[i] = cluster
                    else:
                        lst[i] += ' '+cluster
        return lst

    def pack(self, epsilon=0, coveringOverlap=False):
        """
        merge segments with a gap less than epsilon
        :param epsilon: a int value
        :param coveringOverlap: a boolean value
        """
        
        #index = self.make_index(['show'])
        #lst = list()
        #for show in index:
        #    diar = index[show]
        #    diar.sort(['start'])
        #    i = 0
        #    while i < len(diar.segments) - 1:
        #        if diar.segments[i]['cluster'] == diar.segments[i + 1]['cluster']:
        #            l = Segment.gap(diar.segments[i], diar.segments[i + 1]).duration()
        #            if l <= epsilon:
        #                diar.segments[i]['stop'] = max(diar.segments[i]['stop'],
        #                                               diar.segments[i + 1]['stop'])
        #                del diar.segments[i + 1]
        #            else:
        #                i += 1
        #        else:
        #            i += 1
        #    lst += diar.segments
        #self.segments = lst
        if coveringOverlap:
            index = self.make_index(['show', 'cluster'])
            lst = list()
            for show in index:
                for cluster in index[show]:
                    index[show][cluster].sort(['start'])
                    diar = index[show][cluster]
                    i = 0
                    while i < len(diar.segments) - 1:
                        l = Segment.gap(diar.segments[i], diar.segments[i + 1]).duration()
                        if l <= epsilon:
                            diar.segments[i]['stop'] = max(diar.segments[i]['stop'],
                                                           diar.segments[i + 1]['stop'])
                            del diar.segments[i + 1]
                        else:
                            i += 1
                    lst += diar.segments
            self.segments = lst
        else:
            self.sort(['show', 'start'])
            i = 0
            while i < len(self.segments) - 1:
                if self.segments[i]['show'] == self.segments[i + 1]['show'] and \
                        self.segments[i]['cluster'] == self.segments[i + 1]['cluster'] and \
                        Segment.gap(self.segments[i], self.segments[i + 1]).duration() <= epsilon:
                    self.segments[i]['stop'] = self.segments[i + 1]['stop']
                    del self.segments[i + 1]
                else:
                    i += 1

    def pad(self, epsilon=0):
        """
        Add epsilon frames to the start and stop of each segment
        :param epsilon: the int value to remove
        :return:
        """
        self.sort(['start'])
        i = 0
        if len(self.segments) > 1:
            self.segments[i]['stop'] = max(self.segments[i]['start'], min(max(self.segments[i + 1]['start'] - (epsilon // 2), 0), self.segments[i]['stop'] + epsilon))
        i += 1
        while i < len(self.segments)-1:
            self.segments[i]['start'] = max(self.segments[i - 1]['stop'], self.segments[i]['start'] - epsilon, 0)
            self.segments[i]['stop'] = max(self.segments[i]['start'],min(max(self.segments[i + 1]['start'] - (epsilon // 2), 0), self.segments[i]['stop'] + epsilon))
            i += 1

    def collar(self, epsilon=0, warning=False):
        """
        Apply a collar on each segment. A collar is the no-score zone around
        reference speaker segment boundaries.  (Speaker Diarization output is
        not evaluated within +/- collar seconds of a reference speaker segment
        boundary.)
        :param epsilon: the int value to add
        """
        self.sort(['start'])
        rm = False
        for segment in self.segments:
            segment['stop'] -= epsilon
            segment['start'] += epsilon
            if segment['start'] < 0:
                segment['start'] = 0
            if segment['start'] > segment['stop']:
                segment['start'] = segment['stop']
                rm = True
                if warning:
                    logging.warning('no more segment: '+str(segment['start']-epsilon))
        if rm:
            self.segments = [seg for seg in self.segments if seg.duration() > 0]

    def duration(self):
        """
        :return: t:qhe sum of the segment duration
        """
        l = 0
        for segment in self.segments:
            l += segment.duration()
        return l

    def last_feature_index(self):
        last = 0
        for segment in self.segments:
            if segment['stop'] > last:
                last = segment['stop']
        return last

    def first_feature_index(self):
        if len(self.segments) <= 0:
            return 0
        first = self.segments[0]['start']
        for segment in self.segments:
            if segment['start'] < first:
                first = segment['start']
        return first
    
    @classmethod
    def read_seg(cls, filename, normalize_cluster=False, encoding="utf8"):
        """
        Read a segmentation file
        :param filename: the str input filename
        :param normalize_cluster: normalize the cluster speaker by removing upper
        case and accents
        :return: a diarization object
        """
        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        if not diarization._attributes.exist('gender'):
            diarization.add_attribut(new_attribut='gender', default='U')
        if not diarization._attributes.exist('env'):
            diarization.add_attribut(new_attribut='env', default='U')
        if not diarization._attributes.exist('channel'):
            diarization.add_attribut(new_attribut='channel', default='U')
        try:
            for line in fic: 
                line = re.sub('\s+',' ',line)
                line = line.strip()
                # logging.debug(line)
                if line.startswith('#') or line.startswith(';;'):
                    continue
                # split line into fields
                show, tmp, start, length, gender, channel, environment, name = line.split()
                if normalize_cluster:
                    name = str2str_normalize(name)
                # print(show, tmp, start, length, gender, channel, env, speaker)
                diarization.append(show=show, cluster=name, start=int(start),
                             stop=int(length) + int(start), env=environment,
                             channel=channel,
                             gender=gender)
        except Exception as e:
            logging.error(sys.exc_info()[0])
            # logging.error(line)
        fic.close()
        return diarization

    @classmethod
    def read_ctm(cls, filename, normalize_cluster=False, encoding="utf8"):
        """
        Read a segmentation file
        :param filename: the str input filename
        :param normalize_cluster: normalize the cluster by removing upper case
        and accents
        :return: a diarization object
        """
        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        try:
            for line in fic:
                line = re.sub('\s+',' ',line)
                line = line.strip()
                # logging.debug(line)
                if line.startswith('#') or line.startswith(';;'):
                    continue
                # split line into fields
                show, tmp, start, length, word = line.split()
                if normalize_cluster:
                    word = str2str_normalize(word)
                # print(show, tmp, start, length, gender, channel, env, speaker)
                diarization.append(show=show, cluster=word, start=int(start),
                             stop=int(length) + int(start))
        except Exception as e:
            logging.error(sys.exc_info()[0])
            # logging.error(line)
        fic.close()
        return diarization

    @classmethod
    def read_stm(cls,filename, normalize_cluster=False, encoding="ISO-8859-1"):
        """
        Read a segmentation file
        :param filename: the str input filename
        :param normalize_cluster: normalize the cluster by removing upper case
        and accents
        :return: a diarization object
        """
        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        if not diarization._attributes.exist('gender'):
            diarization.add_attribut(new_attribut='gender', default='U')
        try:
            for line in fic:
                line = re.sub('\s+',' ',line)
                line = line.strip()
                # logging.debug(line)
                if line.startswith('#') or line.startswith(';;'):
                    continue
                # split line into fields
                split = line.split()
                show = split[0]
                loc = split[2]
                if normalize_cluster:
                    loc = str2str_normalize(loc)
                start = int(float(split[3])*100)
                stop = int(float(split[4])*100)
                addon = split[5].replace(">", "").replace("<", "")
                lineBis = re.sub('\s+',' ',addon)
                lineBis = lineBis.strip()
                gender = lineBis.split(",")[2]
                if normalize_cluster:
                    word = str2str_normalize(word)
                # print(show, tmp, start, length, gender, channel, env, speaker)
                if gender == "female":
                    diarization.append(show=show, cluster=loc, start=start,
                             stop=stop,gender="F")
                elif gender == "male":
                    diarization.append(show=show, cluster=loc, start=start,
                             stop=stop,gender="M")
                else:
                    diarization.append(show=show, cluster=loc, start=start,
                             stop=stop)
        except Exception as e:
            logging.error(sys.exc_info()[0])
            logging.error(line)
        fic.close()
        return diarization

    @classmethod
    def read_mdtm(cls, filename, normalize_cluster=False, encoding="utf8"):
        """
        Read a MDTM file
        :param filename: the str input filename
        :param normalize_cluster: normalize the cluster by removing upper case
        and accents
        :return: a diarization object
        """

        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        if not diarization._attributes.exist('gender'):
            diarization.add_attribut(new_attribut='gender', default='U')
        for line in fic:
            line = line.strip()
            line = re.sub('\s+',' ',line)
            logging.debug(line)
            if line.startswith('#') or line.startswith(';;'):
                continue
            # split line into fields
            show, tmp, start_str, length, t, score, gender, cluster = line.split()
            start = int(round(float(start_str)*100, 0))
            stop = start+int(round(float(length)*100, 0))
            if normalize_cluster:
                cluster = str2str_normalize(cluster)
            # print(show, tmp, start, length, gender, channel, env, speaker)
            diarization.append(show=show, cluster=cluster, start=start,
                             stop=stop, gender=gender)
        fic.close()
        return diarization

    @classmethod
    def read_uem(cls, filename, encoding="utf8"):
        """
        Read a UEM file
        :param filename: the str input filename
        :return: a diarization object
        """
        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        if not diarization._attributes.exist('gender'):
            diarization.add_attribut(new_attribut='gender', default='U')
        try:
            name = "uem"
            for line in fic:
                line = re.sub('\s+',' ',line)
                line = line.strip()
                # logging.debug(line)
                if line.startswith('#') or line.startswith(';;'):
                    continue
                # split line into fields
                show, tmp, start_str, stop_str = line.split()
                start = int(round(float(start_str)*100, 0))
                stop = int(round(float(stop_str)*100, 0))
                # stop = start+int(round(float(length)*100, 0))
                diarization.append(show=show, cluster=name, start=start, stop=stop)
        except Exception as e:
            logging.error(sys.exc_info()[0])
            logging.error(line)
        fic.close()
        return diarization

    @classmethod
    def read_rttm(cls, filename, normalize_cluster=False, encoding="utf8"):
        """
        Read rttm file
        :param filename: str input filename
        :param normalize_cluster: normalize the cluster by removing upper case and accents
        :return: a diarization object
        """
        fic = open(filename, 'r', encoding=encoding)
        diarization = Diar()
        if not diarization._attributes.exist('gender'):
            diarization.add_attribut(new_attribut='gender', default='U')
        try:
            for line in fic:
                line = re.sub('\s+',' ',line)
                line = line.strip()
                if line.startswith('#') or line.startswith(';;'):
                    continue
                # split line into fields
                spk, show, tmp0, start_str, length, tmp1, tmp2, cluster, tmp3 = line.split()
                if spk == "SPEAKER":
                    start = int(round(float(start_str)*100, 0))
                    stop = start+int(round(float(length)*100, 0))
                    if normalize_cluster:
                        cluster = str2str_normalize(cluster)
                    diarization.append(show=show, cluster=cluster, start=start, stop=stop)
        except Exception as e:
            logging.error(sys.exc_info()[0])
            logging.error(line)
        fic.close()
        return diarization

    @classmethod
    def to_string_seg(cls, diar, time_float=False):
        """
        transform a diarization into a string
        :param diar: a diarization
        :return: a string
        """
        lst = []
        for segment in diar:
            gender = 'U'
            if diar._attributes.exist('gender'):
                gender = segment['gender']
            env = 'U'
            if diar._attributes.exist('env'):
                env = segment['env']
            channel = 'U'
            if diar._attributes.exist('channel'):
                channel = segment['channel']
            if not time_float:
                lst.append('{:s} 1 {:d} {:d} {:s} {:s} {:s} {:s}\n'.format(
                    segment['show'], segment['start'], segment['stop'] - segment['start'], gender,
                    channel, env, segment['cluster']))
            else:
                lst.append('{:s} 1 {:.2f} {:.2f} {:s} {:s} {:s} {:s}\n'.format(
                    segment['show'], segment['start'], segment['stop'] - segment['start'], gender,
                    channel, env, segment['cluster']))

        return lst

    @classmethod
    def intersection(cls, diarization1, diarization2):
        """
        Compute the intersection between two diarization
        :param diarization1: first diarization
        :param diarization2: second diarization
        :return: a diarization object
        """
        diarization = Diar()
        for segment1 in diarization1:
            for segment2 in diarization2:
                inter = Segment.intersection(segment1, segment2)
                if inter is not None :
                    diarization.append_seg(inter)
        return diarization

    @classmethod
    def write_seg(cls, filename, diarization):
        """
        Write diarization to a segmentation file
        :param filename: the str output filename
        :param diarization: the diarization to write

        """
        diarization.sort(['show', 'start'])
        fic = open(filename, 'w', encoding="utf8")
        for line in Diar.to_string_seg(diarization):
            fic.write(line)
        fic.close()

    @classmethod
    def write_lbl(cls, diarization, label_dir='', label_file_extension='.lbl'):
        """
        Write diarization to label file
        :param diarization: the diarization to write
        :param label_dir: the string directory of the ouput filename
        :param label_file_extension: the string extension of the output filename

        """
        diarization.sort(['show', 'start'])
        old_show = ''
        fic = None
        for segment in diarization:
            if old_show != segment['show']:
                if fic is not None:
                    fic.close()
                filename = os.path.join(label_dir, segment['show']
                                        + label_file_extension)
                fic = open(filename, 'w')
                old_show = segment['show']
            fic.write('{:d} {:d} {:s}\n'.format(
                segment['start'], segment['stop'], segment['cluster']))
        fic.close()

class Segment(list):
    """
    Class to store the segment informations.


    :attr _attributes: is the list of attribut names
    :attr data: the data associated to each attribut
    """
    def __init__(self, data, attributes):
        """
        Called after the instance has been created (by __new__()), but before it is returned to the caller.
        :param data: copy the row data
        :param attributes: the names of the attributs

        """
        list.__init__(self)
        self._attributes = attributes
        for item in data:
            self.append(item)

    def _get_attr(self, attr_name):
        """
        Called to implement evaluation of self[attr_name].
        :param attr_name: a string
        :return: the value
        """
        return self[self._attributes[attr_name]]

    def _set_attr(self, attr_name, value):
        """
        Called to implement assignment to self[attr_name].

        :param attr_name: a str
        :param value: the value to set

        """
        self[self._attributes[attr_name]] = value

    def __getitem__(self, index):
        """
        Called to implement evaluation of self[index].
        :param index: a int
        :return: the value
        """

        if isinstance(index, str):
            return self._get_attr(index)
        else:
            return list.__getitem__(self, index)

    def __setitem__(self, index, value):
        """
        Called to implement assignment to self[index].
        :param index: a int
        :param value: the value to set
        :return: the item
        """
        if isinstance(index, str):
            return self._set_attr(index, value)
        else:
            return list.__setitem__(self, index, value)

    def __eq__(self, segment): # real signature unknown
        if segment is not None:
            l = len(segment)
            if l != len(self):
                return False

            for i in range(l):
                if self[i] != segment[i]:
                    return False
            return True
        else:
            return False

    def __ne__(self, segment): # real signature unknown
        return not self.__eq__(segment)

    def duration(self):
        """
        :return: the duration of the segment
        """

        return self['stop'] - self['start']

    def seg_features(self, features):
        """
        Given a FeatureServer, returns a list of feature index corresponding to
        the segment.

        :param features: a FeatureServer
        :return: a list of int
        """
        return features[self['start']:self['stop'], :]

    @classmethod
    def gap(cls, segment1, segment2):
        """
        Returns the inter segment gap between 2 segments.
        :param segment1: a Segment object
        :param segment2: a Segment object
        :return: a Segment object

        Examples
        --------
        >>> from s4d.diarization import Diar, Segment
        >>> diarization=Diar()
        >>> diarization.append(show='empty', start=0, stop=100, cluster='spk1')
        >>> diarization.append(show='empty', start=50, stop=150, cluster='spk2')
        >>> s = Segment.intersection(diarization[0], diarization[1])
        >>> s
        ['empty', 'spk1', 'speaker', 100, 50]
        >>> s.duration()
        - 50
        >>> diarization.append(show='empty', start=200, stop=250, cluster='spk1')
        >>> Segment.gap(diarization[0], diarization[2])
        ['empty', 'spk1', 'speaker', 100, 200]


        """
        if segment1['show'] != segment2['show']:
            raise Exception('not the same show')
        segment = Segment(segment1, segment1._attributes)
        segment['start'] = segment1['stop']
        segment['stop'] = segment2['start']
        return segment
    @classmethod
    def split(cls, segment1, segment2):
        inter = cls.intersection(segment1, segment2)
        diff = cls.diff(segment1, segment2)

    @classmethod
    def intersection(cls, segment1, segment2):
        """
        Intersection between 2 segments. Return None if the intersection is empty.
        :param segment1: a Segment object
        :param segment2: a Segment object
        :return: a Segment object

        Examples
        --------
        >>> from s4d.diarization import Diar, Segment
        >>> diarization=Diar()
        >>> diarization.append(show='empty', start=0, stop=100, cluster='spk1')
        >>> diarization.append(show='empty', start=50, stop=150, cluster='spk2')
        >>> Segment.intersection(diarization[0], diarization[1])
        ['empty', 'spk1 / spk2', 'speaker', 50, 100]
        >>> diarization.append(show='empty', start=50, stop=75, cluster='spk1')
        >>> Segment.intersection(diarization[0], diarization[2])
        ['empty', 'spk1 / spk1', 'speaker', 50, 75]
        >>> diarization.append(show='empty', start=200, stop=250, cluster='spk1')
        >>> s = Segment.intersection(diarization[0], diarization[3])
        >>> s is None

        True

        """
        if segment1['show'] != segment2['show']:
            raise Exception(
                'not the same show ' + segment1['show'] + ' != ' + segment2['show'])
        segment = Segment(segment1, segment1._attributes)
        segment['cluster'] += '##' + segment2['cluster']
        segment['start'] = max(segment1['start'], segment2['start'])
        segment['stop'] = min(segment1['stop'], segment2['stop'])
        if segment.duration() > 0:
            return segment
        return None

    @classmethod
    def diff(cls, segment1, segment2):
        """
        The difference between the two segments. Returns one or two segments and the
        source of the new segment: 1 means segment1, 2 means segment2.

        :param segment1: a Segment object
        :param segment2: a Segment object
        :return: a list of segments
        """
        lst_row = list()

        if segment1['show'] != segment2['show']:
            return lst_row
        start = min(segment1['start'], segment2['start'])
        stop = max(segment1['start'], segment2['start'])
        if stop - start > 0:
            if start == segment1['start']:
                segment = copy.deepcopy(segment1)
            else:
                segment = copy.deepcopy(segment2)
            segment['stop'] = stop
            lst_row.append(segment)

        start = min(segment1['stop'], segment2['stop'])
        stop = max(segment1['stop'], segment2['stop'])
        if stop - start > 0:
            if stop == segment1['stop']:
                segment = copy.deepcopy(segment1)
            else:
                segment = copy.deepcopy(segment2)
            segment['start'] = start
            lst_row.append(segment)
        return lst_row

    @classmethod
    def union(cls, segment1, segment2):
        """
        Union between 2 segments.
        :param segment1: a Segment object
        :param segment2: a Segment object
        :return: a Segment object

        Examples
        --------
        >>> from s4d.diarization import Diar, Segment
        >>> diarization=Diar()
        >>> diarization.append(show='empty', start=0, stop=100, cluster='spk1')
        >>> diarization.append(show='empty', start=50, stop=150, cluster='spk2')
        >>> Segment.union(diarization[0], diarization[1])
        ['empty', 'spk1', 'speaker', 0, 150]
        >>> diarization.append(show='empty', start=50, stop=75, cluster='spk1')
        >>> Segment.union(diarization[0], diarization[2])
        ['empty', 'spk1', 'speaker', 0, 100]
        >>> diarization.append(show='empty', start=200, stop=250, cluster='spk1')
        >>> Segment.union(diarization[0], diarization[3])

        ['empty', 'spk1', 'speaker', 0, 250]

        True

        """
        if segment1['show'] != segment2['show']:
            raise Exception('not the same show ' +
                            segment1['show'] + ' != ' + segment2['show'])
        segment = Segment(segment1, segment1._attributes)
        segment['start'] = min(segment1['start'], segment2['start'])
        segment['stop'] = max(segment1['stop'], segment2['stop'])
        return segment


class Index(dict):
    """Implementation of perl's autovivification feature.
    Thanks to http://stackoverflow.com/questions/651794/whats-the-best-way-to-initialize-a-dict-of-dicts-in-python
    """

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            value = self[item] = type(self)()
            return value


class AttributeNames:
    """
    Class AttributeNames defines a list of column names
    """
    def __init__(self):
        self.names = dict
        self.defaults = list

    def index_of(self, name):
        """
        :param name: the speaker of the column
        :return: the position (a int) of the speaker
        """
        return self.names[name]

    def exist(self, name):
        """
        Test the existing value of speaker in the list of column names
        :param name: the speaker of the column
        :return: a boolean
        """
        return name in self.names

    def initialize(self, names, defaults):
        """
        Initialaze AttributeNames object
        :param names: a list of column names
        :param defaults: the list of default values

        """
        self.names = names
        self.defaults = defaults

    def __getitem__(self, index):
        """
        Called to implement evaluation of self[index].
        :param index: a int
        :return: the value
        """
        return self.names[index]

    def __setitem__(self, index, value):
        """
        Called to implement assignment to self[index].
        :param index: a int
        :param value: the value to set
        :return: the item
        """
        return self.names.__setitem__(index, value)

    def __iter__(self):
        """
        This method is called when an iterator is required for a container.
        :return: a iterator on the column names
        """
        return self.names.__iter__()

    def __len__(self):
        """
        Get the number of column names
        :return:
        """
        return self.defaults.__len__()

    def sorted(self):
        """
        sort the column names
        :return:
        """
        return sorted(self.names.items(), key=lambda x: x[1])

    def add(self, name, default=''):
        """
        a a column speaker
        :param name: a str
        :param default: the default value

        """
        if name in self.names:
            raise Exception('This attribut exits : ')
        else:
            self[name] = len(self.defaults)
            self.defaults.append(default)

    def delete(self, name):
        """
        remove a column and the default value  given it column speaker
        :param name:

        """
        if name not in self.names:
            raise Exception("This attribut don't exits : " + name)
        else:
            i = self[name]
            del self.defaults[i]
            del self.names[name]
            for k in self.names:
                if self.names[k] > i:
                    self.names[k] -= 1


def rolling_window(a, window):
    """
    Make an ndarray with a rolling window of the last dimension

    Examples
    --------
    >>> x=numpy.arange(10).reshape((2,5))
    >>> rolling_window(x, 3)
    array([[[0, 1, 2], [1, 2, 3], [2, 3, 4]],
           [[5, 6, 7], [6, 7, 8], [7, 8, 9]]])

    Calculate rolling mean of last dimension:
    >>> numpy.mean(rolling_window(x, 3), -1)
    array([[ 1.,  2.,  3.],
           [ 6.,  7.,  8.]])

    :param a: Array to add rolling window to
    :param window: Size of rolling window
    :return: Array that is a view of the original array with a added dimension
    of size w.
    """

    if window < 1:
        raise (ValueError, "`window` must be at least 1.")
    if window > a.shape[-1]:
        raise (ValueError, "`window` is too long.")
    shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return numpy.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)






# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2021 Anthony Larcher
"""

__license__ = "LGPL"
__author__ = "Anthony Larcher"
__copyright__ = "Copyright 2015-2021 Anthony Larcher and Sylvain Meignier"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reStructuredText'

import numpy
import pathlib
import random
import scipy
import sidekit
import soundfile
import torch
import tqdm
import re
import yaml

from ..diar import Diar
from pathlib import Path
from sidekit.nnet.xsets import PreEmphasis
from sidekit.nnet.xsets import MFCC
from sidekit.nnet.xsets import CMVN
from sidekit.nnet.xsets import FrequencyMask
from sidekit.nnet.xsets import TemporalMask
from torch.utils.data import Dataset
from torchvision import transforms
from collections import namedtuple

#Segment = namedtuple('Segment', ['show', 'start_time', 'end_time'])

def read_ctm(filename, normalize_cluster=False, encoding="utf8"):
    """
    Read a segmentation file
    :param filename: the str input filename
    :return: a diarization object
    :param normalize_cluster: normalize the cluster by removing upper case
    """
    fic = open(filename, 'r', encoding=encoding)
    diarization = Diar()

    for line in fic:
        line = re.sub('\s+',' ',line)
        line = line.strip()
        if line.startswith('#') or line.startswith(';;'):
            continue
            # split line into fields
        show=line.split()[0]
        start= int(float(line.split()[2])*1000)/10.
        stop= start + int(float(line.split()[3])*1000)/10.
        word= line.split()[4]

    if normalize_cluster:
        word = str2str_normalize(word)

    diarization.append(show=show, cluster=word, start=float(start), stop=stop)
    fic.close()

    return diarization


def overlapping(seg1,seg2):
    seg1_start,seg1_stop=seg1
    seg2_start,seg2_stop=seg2
    if seg1_start <= seg2_start:
        return seg1_stop > seg2_start
    else:
        return seg2_stop > seg1_start


def framing(sig, win_size, win_shift=1, context=(0, 0), pad='zeros'):
    """
    :param sig: input signal, can be mono or multi dimensional
    :param win_size: size of the window in term of samples
    :param win_shift: shift of the sliding window in terme of samples
    :param context: tuple of left and right context
    :param pad: can be zeros or edge
    """
    dsize = sig.dtype.itemsize
    if sig.ndim == 1:
        sig = sig[:, numpy.newaxis]
    # Manage padding
    c = (context, ) + (sig.ndim - 1) * ((0, 0), )
    _win_size = win_size + sum(context)
    shape = (int((sig.shape[0] - win_size) / win_shift) + 1, 1, _win_size, sig.shape[1])
    strides = tuple(map(lambda x: x * dsize, [win_shift * sig.shape[1], 1, sig.shape[1], 1]))
    return numpy.lib.stride_tricks.as_strided(sig,
                                           shape=shape,
                                           strides=strides).squeeze()

def load_wav_segment(wav_file_name, idx, duration, seg_shift, framerate=16000):
    """

    :param wav_file_name:
    :param idx:
    :param duration:
    :param seg_shift:
    :param framerate:
    :return:
    """
    # Load waveform
    signal = sidekit.frontend.io.read_audio(wav_file_name, framerate)[0]
    tmp = framing(signal,
                  int(framerate * duration),
                  win_shift=int(framerate * seg_shift),
                  context=(0, 0),
                  pad='zeros')
    return tmp[idx], len(signal)


def mdtm_to_label(mdtm_filename,
                  start_time,
                  stop_time,
                  sample_number,
                  speaker_dict,
                  is_uem=False,
                  is_ctm=False):
    """

    :param mdtm_filename:
    :param start_time:
    :param stop_time:
    :param sample_number:
    :param speaker_dict:
    :return:
    """
    if is_uem:
        diarization = Diar.read_uem(mdtm_filename)
    elif is_ctm:
        diarization = read_ctm(mdtm_filename)
    else:
        diarization = Diar.read_mdtm(mdtm_filename)
    diarization.sort(['show', 'start'])

    # When one segment starts just the frame after the previous one ends, o
    # we replace the time of the start by the time of the previous stop to avoid artificial holes
    previous_stop = 0
    for ii, seg in enumerate(diarization.segments):
        if ii == 0:
            previous_stop = seg['stop']
        else:
            if seg['start'] == diarization.segments[ii - 1]['stop'] + 1:
                diarization.segments[ii]['start'] = diarization.segments[ii - 1]['stop']

    # Create the empty labels
    label = []

    # Compute the time stamp of each sample
    time_stamps = numpy.zeros(sample_number, dtype=numpy.float32)
    period = (stop_time - start_time) / sample_number
    for t in range(sample_number):
        time_stamps[t] = start_time + (2 * t + 1) * period / 2

    for idx, time in enumerate(time_stamps):
        lbls = []
        for seg in diarization.segments:
            if seg['start'] / 100. <= time <= seg['stop'] / 100.:
                if speaker_dict is None:
                    lbls.append("1")
                else:
                    lbls.append(speaker_dict[seg['cluster']])

        if len(lbls) > 0:
            label.append(lbls)
        else:
            label.append([])

    if is_uem:
        tmp_label = []
        for lbl in label:
            tmp_label.append(len(lbl))
        label = tmp_label

    return label


def get_segment_label(label,
                      seg_idx,
                      mode,
                      duration,
                      framerate,
                      seg_shift,
                      collar_duration,
                      filter_type="gate"):
    """

    :param label:
    :param seg_idx:
    :param mode:
    :param duration:
    :param framerate:
    :param seg_shift:
    :param collar_duration:
    :param filter_type:
    :return:
    """

    # Create labels with Diracs at every speaker change detection
    spk_change = numpy.zeros(label.shape, dtype=int)
    spk_change[:-1] = label[:-1] ^ label[1:]
    spk_change = numpy.not_equal(spk_change, numpy.zeros(label.shape, dtype=int))

    # depending of the mode, generates the labels and select the segments
    if mode == "vad":
        output_label = (label > 0.5).astype(numpy.long)

    elif mode == "spk_turn":
        # Apply convolution to replace diracs by a chosen shape (gate or triangle)
        filter_sample = collar_duration * framerate * 2 + 1
        conv_filt = numpy.ones(filter_sample)
        if filter_type == "triangle":
            conv_filt = scipy.signal.triang(filter_sample)
        output_label = numpy.convolve(conv_filt, spk_change, mode='same')

    elif mode == "overlap":
        output_label = (label > 0.5).astype(numpy.long)

    else:
        raise ValueError("mode parameter must be 'vad', 'spk_turn' or 'overlap', 'resegmentation'")

    # Create segments with overlap
    segment_label = framing(output_label,
                  int(framerate * duration),
                  win_shift=int(framerate * seg_shift),
                  context=(0, 0),
                  pad='zeros')

    return segment_label[seg_idx]


def process_segment_label(label,
                          mode,
                          framerate,
                          collar_duration,
                          filter_type="gate"):
    """

    :param label:
    :param seg_idx:
    :param mode:
    :param duration:
    :param framerate:
    :param seg_shift:
    :param collar_duration:
    :param filter_type:
    :return:
    """

    # depending of the mode, generates the labels and select the segments
    if mode == "vad":
        output_label = numpy.array([len(a) > 0 for a in label]).astype(numpy.long)

    elif mode == "spk_turn":

        tmp_label = []
        for a in label:
            if len(a) == 0:
                tmp_label.append(0)
            elif len(a) == 1:
                tmp_label.append(a[0])
            else:
                tmp_label.append(sum(a) * 1000)

        label = numpy.array(tmp_label)

        # Create labels with Diracs at every speaker change detection
        spk_change = numpy.zeros(label.shape, dtype=int)
        spk_change[:-1] = label[:-1] ^ label[1:]
        spk_change = numpy.not_equal(spk_change, numpy.zeros(label.shape, dtype=int))

        # Apply convolution to replace diracs by a chosen shape (gate or triangle)
        filter_sample = int(collar_duration * framerate * 2 + 1)

        conv_filt = numpy.ones(filter_sample)
        if filter_type == "triangle":
            conv_filt = scipy.signal.triang(filter_sample)
        output_label = numpy.convolve(conv_filt, spk_change, mode='same')

    elif mode == "overlap":
        label = numpy.array([len(a) for a in label]).astype(numpy.long)

        # For the moment, we just consider two classes: overlap / no-overlap
        # in the future we might want to classify according to the number of speaker speaking at the same time
        output_label = (label > 1).astype(numpy.long)

    elif mode == "resegmentation":
        tmp_label = []
        for a in label:
            if len(a) == 0:
                tmp_label.append(0)
            elif len(a) == 1:
                tmp_label.append(a[0])
            else:
                tmp_label.append(sum(a) * 1000)

        output_label = numpy.array(tmp_label)

    else:
        raise ValueError("mode parameter must be 'vad', 'spk_turn' or 'overlap', 'resegmentation'")

    return output_label


def seqSplit(mdtm_dir,
             wav_dir,
             mode='vad',
             duration=2.,
             file_list=None):
    """
    
    :param mdtm_dir:
    :param mode: can be 'vad' or 'spk_turn'
    :param duration: 
    :return: 
    """
    segment_list = Diar()
    speaker_dict = dict()
    idx = 0

    # Get the list of shows to process
    if file_list is None:
        # For each MDTM
        show_list = []
        for mdtm_file in tqdm.tqdm(pathlib.Path(mdtm_dir).glob('*.mdtm')):
            show_list.append(str(mdtm_file)[len(mdtm_dir):].split(".")[0])
    else:
        with open(file_list, 'r') as fh:
            tmp = fh.readlines()
            show_list = [l.rstrip() for l in tmp if not l == '']

    #for mdtm_file in tqdm.tqdm(pathlib.Path(mdtm_dir).glob('*.mdtm')):
    for show_name in show_list:

        # Load MDTM file
        ref = Diar.read_mdtm(mdtm_dir + show_name + ".mdtm")
        ref.sort()
        last_stop = ref.segments[-1]["stop"]

        #showName = str(mdtm_file)[len(mdtm_dir):].split(".")[0]

        if mode == 'vad':
            _stop = float(ref.segments[-1]["stop"]) / 100.
            _start = float(ref.segments[0]["start"]) / 100.

            while _start + 2 * duration < _stop:
                segment_list.append(show=show_name,
                                    cluster="",
                                    start=_start,
                                    stop=_start + 2 * duration)
                _start += duration / 4.

        elif mode == "spk_turn" or mode == "overlap":
            # Get the borders of the segments (not the start of the first and not the end of the last

            # Check the length of audio
            nfo = soundfile.info(wav_dir + show_name + ".wav")

            # For each border time B get a segment between B - duration and B + duration
            # in which we will pick up randomly later
            for idx, seg in enumerate(ref.segments):

                if seg["start"] / 100. > duration and seg["start"] / 100. + duration < nfo.duration:
                    segment_list.append(show=seg['show'],
                                        cluster="",
                                        start=float(seg["start"]) / 100. - duration,
                                        stop=float(seg["start"]) / 100. + duration)
                if seg["stop"] / 100. > duration and seg["stop"] / 100. + duration < nfo.duration:
                    segment_list.append(show=seg['show'],
                                        cluster="",
                                        start=float(seg["stop"]) / 100. - duration,
                                        stop=float(seg["stop"]) / 100. + duration)

        # Get list of unique speakers
        speakers = ref.unique('cluster')
        for spk in speakers:
            if not spk in speaker_dict:
                speaker_dict[spk] =  idx
                idx += 1

    return segment_list, speaker_dict


class SeqSet(Dataset):
    """
    Object creates a dataset for sequence to sequence training
    """
    def __init__(self,
                 dataset_yaml,
                 wav_dir,
                 mdtm_dir,
                 mode,
                 segment_list=None,
                 speaker_dict=None,
                 filter_type="gate",
                 collar_duration=0.1,
                 audio_framerate=16000,
                 output_framerate=100,
                 output_sample_number=None):
        """

        :param wav_dir:
        :param mdtm_dir:
        :param mode:
        :param duration:
        :param filter_type:
        :param collar_duration:
        :param audio_framerate:
        :param output_framerate:
        :param transform_pipeline:
        """

        self.wav_dir = wav_dir
        self.mdtm_dir = mdtm_dir
        self.mode = mode
        self.filter_type = filter_type
        self.collar_duration = collar_duration
        self.audio_framerate = audio_framerate
        self.output_framerate = output_framerate
        self.output_sample_number = output_sample_number

        self.duration = dataset_yaml["train"]["duration"]

        if mode == "train":
            self.transformation  = dataset_yaml["train"]["transformation"]

        else:
            self.transformation  = dataset_yaml["eval"]["transformation"]


        self.segment_list = segment_list
        self.speaker_dict = speaker_dict
        self.len = len(segment_list)

        _transform = []
        if self.transformation["pipeline"] != '' and self.transformation["pipeline"] is not None:


            self.add_noise = numpy.zeros(self.len, dtype=bool)
            self.add_reverb = numpy.zeros(self.len, dtype=bool)
            self.spec_aug = numpy.zeros(self.len, dtype=bool)
            self.temp_aug = numpy.zeros(self.len, dtype=bool)

            trans = self.transformation["pipeline"].split(',')
            for t in trans:
                if 'PreEmphasis' in t:
                    _transform.append(PreEmphasis())

                if 'add_noise' in t:
                    self.add_noise[:int(self.len * self.transformation["noise_file_ratio"])] = 1
                    numpy.random.shuffle(self.add_noise)
                    _transform.append(AddNoise(noise_db_csv=self.transformation["noise_db_csv"],
                                               snr_min_max=self.transformation["noise_snr"],
                                               noise_root_path=self.transformation["noise_root_db"]))

                if 'add_reverb' in t:
                    has_pyroom = True
                    try:
                        import pyroomacoustics
                    except ImportError:
                        has_pyroom = False
                    if has_pyroom:
                        self.add_reverb[:int(self.len * self.transformation["reverb_file_ratio"])] = 1
                        numpy.random.shuffle(self.add_reverb)

                        _transform.append(AddReverb(depth=self.transformation["reverb_depth"],
                                                    width=self.transformation["reverb_width"],
                                                    height=self.transformation["reverb_height"],
                                                    absorption=self.transformation["reverb_absorption"],
                                                    noise=None,
                                                    snr=self.transformation["reverb_snr"]))

                if 'MFCC' in t:
                    _transform.append(MFCC())

                if "CMVN" in t:
                    _transform.append(CMVN())

                if "FrequencyMask" in t:
                    a = int(t.split('-')[0].split('(')[1])
                    b = int(t.split('-')[1].split(')')[0])
                    _transform.append(FrequencyMask(a, b))

                if "TemporalMask" in t:
                    a = int(t.split("(")[1].split(")")[0])
                    _transform.append(TemporalMask(a))

        self.transforms = transforms.Compose(_transform)

        if segment_list is None and speaker_dict is None:
            segment_list, speaker_dict = seqSplit(mdtm_dir=self.mdtm_dir,
                                                  duration=self.duration)
    def __getitem__(self, index):
        """
        On renvoie un segment wavform brut mais il faut que les labels soient échantillonés à la bonne fréquence
        (trames)
        :param index:
        :return:
        """
        # Get segment info to load from
        seg = self.segment_list[index]

        # Randomly pick an audio chunk within the current segment
        start = random.uniform(seg["start"], seg["start"] + self.duration)

        sig, _ = soundfile.read(self.wav_dir + seg["show"] + ".wav",
                                start=int(start * self.audio_framerate),
                                stop=int((start + self.duration) * self.audio_framerate)
                                )
        sig += 0.0001 * numpy.random.randn(sig.shape[0])

        if self.transformation:
            sig, speaker_idx, _, __, _t, _s = self.transforms((sig, None,  None, None, None, None))


        if self.output_sample_number is None:
            tmp_label = mdtm_to_label(mdtm_filename=self.mdtm_dir + seg["show"] + ".mdtm",
                                      start_time=start,
                                      stop_time=start + self.duration,
                                      sample_number=sig.shape[-1],
                                      speaker_dict=self.speaker_dict)

            label = process_segment_label(label=tmp_label,
                                          mode=self.mode,
                                          framerate=self.output_framerate,
                                          collar_duration=None)

        else:
            tmp_label = mdtm_to_label(mdtm_filename=self.mdtm_dir + seg["show"] + ".mdtm",
                                      start_time=start,
                                      stop_time=start + self.duration,
                                      sample_number=self.output_sample_number,
                                      speaker_dict=self.speaker_dict)

            label = process_segment_label(label=tmp_label,
                                          mode=self.mode,
                                          framerate=self.duration / float(self.output_sample_number),
                                          collar_duration=None)


        return torch.from_numpy(sig.T).type(torch.FloatTensor), torch.from_numpy(label.astype('long'))

    def __len__(self):
        return self.len


def create_train_val_seqtoseq(dataset_yaml):
    """

    :param self:
    :param wav_dir:
    :param mdtm_dir:
    :param mode:
    :param segment_list
    :param speaker_dict:
    :param filter_type:
    :param collar_duration:
    :param audio_framerate:
    :param output_framerate:
    :param transform_pipeline:
    :return:
    """
    with open(dataset_yaml, "r") as fh:
        dataset_params = yaml.load(fh, Loader=yaml.FullLoader)

    torch.manual_seed(dataset_params['seed'])

    # Read all MDTM files and ouptut a list of segments with minimum duration as well as a speaker dictionary
    segment_list, speaker_dict = seqSplit(mdtm_dir=dataset_params["mdtm_dir"],
                                          wav_dir=dataset_params["wav_dir"],
                                          duration=dataset_params["train"]["duration"],
                                          file_list=dataset_params["train"]["file_list"])

    split_idx = numpy.random.choice([True, False],
                                    size=(len(segment_list),),
                                    p=[1 - dataset_params["validation_ratio"], dataset_params["validation_ratio"]])
    segment_list_train = Diar.copy_structure(segment_list)
    segment_list_val = Diar.copy_structure(segment_list)
    for idx, seg in enumerate(segment_list.segments):
        if split_idx[idx]:
            segment_list_train.append_seg(seg)
        else:
            segment_list_val.append_seg(seg)

    # Split the list of segment between training and validation sets
    train_set = SeqSet(dataset_params,
                       wav_dir=dataset_params["wav_dir"],
                       mdtm_dir=dataset_params["mdtm_dir"],
                       mode=dataset_params["mode"],
                       segment_list=segment_list_train,
                       speaker_dict=speaker_dict,
                       filter_type=dataset_params["filter_type"],
                       collar_duration=dataset_params["collar_duration"],
                       audio_framerate=dataset_params["sample_rate"],
                       output_framerate=dataset_params["output_rate"],
                       output_sample_number=dataset_params["output_sample_number"])

    validation_set = SeqSet(dataset_params,
                            wav_dir=dataset_params["wav_dir"],
                            mdtm_dir=dataset_params["mdtm_dir"],
                            mode=dataset_params["mode"],
                            segment_list=segment_list_val,
                            speaker_dict=speaker_dict,
                            filter_type=dataset_params["filter_type"],
                            collar_duration=dataset_params["collar_duration"],
                            audio_framerate=dataset_params["sample_rate"],
                            output_framerate=dataset_params["output_rate"],
                            output_sample_number=dataset_params["output_sample_number"])

    return train_set, validation_set


def seqSplit_sliding_window(show,
                            mdtm_fn,
                            wav_fn,
                            duration=3.2,
                            speaker_dict=None,
                            shift=2.4):
    """

    :param mdtm_fn:
    :param wav_fn:
    :param uem_fn:
    :param duration:
    :param audio_framerate:
    :param shift:
    :return:
    """

    segment_list = Diar()
    if speaker_dict is None:
        speaker_dict = dict()

    # Load ref file
    ref = Diar.read_mdtm(mdtm_fn)

    # Get list of unique speakers
    idx = 0
    speakers = ref.unique('cluster')
    for spk in speakers:
        if not spk in speaker_dict:
            speaker_dict[spk] =  idx
            idx += 1

    # Check the length of audio
    nfo = soundfile.info(wav_fn)

    start = 0.
    while (start + duration < nfo.duration):
        segment_list.append(show=show,
                            cluster="",
                            start=start,
                            stop=start + duration)
        start += shift

    return segment_list, speaker_dict


class SeqSetSingle(Dataset):
    """
    Object creates a dataset for sequence to sequence training
    """
    def __init__(self,
                 show,
                 wav_fn,
                 mdtm_fn,
                 mode,
                 audio_framerate=16000,
                 output_framerate=100,
                 speaker_dict=None,
                 duration=3.2,
                 shift=2.4,
                 transform_pipeline=""):
        """

        :param wav_dir:
        :param mdtm_dir:
        :param mode:
        :param duration:
        :param filter_type:
        :param collar_duration:
        :param audio_framerate:
        :param output_framerate:
        :param transform_pipeline:
        """

        self.wav_fn = wav_fn
        self.mdtm_fn = mdtm_fn
        self.mode = mode
        self.audio_framerate = audio_framerate
        self.output_framerate = output_framerate
        self.speaker_dict = speaker_dict
        self.duration = duration
        self.shift = shift
        self.transform_pipeline = transform_pipeline

        _transform = []
        if not self.transform_pipeline == '':
            trans = self.transform_pipeline.split(',')
            for t in trans:
                if 'PreEmphasis' in t:
                    _transform.append(PreEmphasis())
                if 'MFCC' in t:
                    _transform.append(MFCC())
                if "CMVN" in t:
                    _transform.append(CMVN())
                if "FrequencyMask" in t:
                    a = int(t.split('-')[0].split('(')[1])
                    b = int(t.split('-')[1].split(')')[0])
                    _transform.append(FrequencyMask(a, b))
                if "TemporalMask" in t:
                    a = int(t.split("(")[1].split(")")[0])
                    _transform.append(TemporalMask(a))

        self.transforms = transforms.Compose(_transform)

        segment_list, speaker_dict = seqSplit_sliding_window(show,
                                                             mdtm_fn=self.mdtm_fn,
                                                             wav_fn=self.wav_fn,
                                                             duration=self.duration,
                                                             speaker_dict=self.speaker_dict,
                                                             shift=self.shift)

        self.segment_list = segment_list
        self.speaker_dict = speaker_dict
        self.len = len(self.segment_list)

    def __getitem__(self, index):
        """
        On renvoie un segment wavform brut mais il faut que les labels soient échantillonés à la bonne fréquence
        (trames)
        :param index:
        :return:
        """
        # Get segment info to load from
        seg = self.segment_list[index]

        sig, _ = soundfile.read(self.wav_fn,
                                start=int(seg["start"] * self.audio_framerate),
                                stop=int((seg["start"] + self.duration) * self.audio_framerate))

        sig += 0.0001 * numpy.random.randn(sig.shape[0])

        if self.transform_pipeline:
            sig, speaker_idx, _, __, _t, _s = self.transforms((sig, None,  None, None, None, None))


        tmp_label = mdtm_to_label(mdtm_filename=self.mdtm_fn,
                                  start_time=(seg["start"]/self.audio_framerate)*100,
                                  stop_time=(seg["stop"]/self.audio_framerate)*100,
                                  sample_number=sig.shape[1],
                                  speaker_dict=self.speaker_dict)

        label = process_segment_label(label=tmp_label,
                                      mode=self.mode,
                                      framerate=self.output_framerate,
                                      collar_duration=None,
                                      filter_type=None)

        return index, torch.from_numpy(sig.T).type(torch.FloatTensor), torch.from_numpy(label.astype('long'))

    def __len__(self):
        return self.len



# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2021 Sylvain Meignier
"""

import pyaudio


class AudioPlayer:
    """
    Player implemented with PyAudio

    http://people.csail.mit.edu/hubert/pyaudio/

    Mac OS X:

      brew install portaudio
      pip install http://people.csail.mit.edu/hubert/pyaudio/packages/pyaudio-0.2.8.tar.gz
    """
    def __init__(self, wav):
        """

        :param wav:
        """
        self.p = pyaudio.PyAudio()
        self.pos = 0
        self.stream = None
        self._open(wav)

    def callback(self, in_data, frame_count, time_info, status):
        """

        :param in_data:
        :param frame_count:
        :param time_info:
        :param status:
        :return:
        """
        data = self.wf.readframes(frame_count)
        self.pos += frame_count
        return (data, pyaudio.paContinue)

    def _open(self, wav):
        """

        :param wav:
        :return:
        """
        self.wf = wav
        self.stream = self.p.open(format=self.p.get_format_from_width(self.wf.getsampwidth()),
                channels = self.wf.getnchannels(),
                rate = self.wf.getframerate(),
                output=True,
                stream_callback=self.callback)
        self.pause()

    def play(self):
        """

        :return:
        """
        self.stream.start_stream()

    def pause(self):
        """

        :return:
        """
        self.stream.stop_stream()

    def seek(self, seconds = 0.0):
        """

        :param seconds:
        :return:
        """
        sec = seconds * self.wf.getframerate()
        self.pos = int(sec)
        self.wf.setpos(int(sec))

    def time(self):
        """

        :return:
        """
        return float(self.pos)/self.wf.getframerate()

    def playing(self):
        """

        :return:
        """
        return self.stream.is_active()

    def close(self):
        """

        :return:
        """
        self.stream.close()
        self.wf.close()
        self.p.terminate()




# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2021 Sylvain Meignier
"""

import logging
import numpy as np


class GaussFull:
    """
    Class to model a speaker by a gaussian with full covariance
    """
    def __init__(self, name, dim):
        self.logger = logging.getLogger(__name__)
        self.name = name
        self.count = 0
        self.dim = dim
        self.stat1 = np.zeros(dim)
        self.stat2 = np.zeros((dim, dim))
        self.cov_log_det = 0;
        self.mu = None
        self.mu_dot = np.NAN
        self.cov = None
        self.partial_bic = np.NaN

    def add(self, features):
        """
        Accumulate statistics for *features*
        :param features: numpy.ndarray

        """
        self.count += features.shape[0]  # add the number of features
        self.stat1 += features.sum(axis=0)
        self.stat2 += np.dot(features.T, features)

    def _cov_log_det(self):
        """
        Compute the log det of the covariance matrix
        :return:  float
        """
        s, d = np.linalg.slogdet(self.cov)
        return d
        # cov_chol, lower = scipy.linalg.cho_factor(self.cov)
        # return 2.0 * numpy.sum(numpy.log(numpy.diagonal(cov_chol)))

    def compute(self):
        """
        Compute the mean and covariance according the statistique, the log det
         of the covariance and the partial BIC :math:`PBIC`.

        :math:`PBIC_{x}  = \\frac{n_x}{2} \\log|\\Sigma_x|`

        """
        self.mu = self.stat1 / self.count
        tmp = self.mu[:, np.newaxis]
        self.mu_dot = np.dot(tmp, tmp.T)
        self.cov = self.stat2 / self.count - self.mu_dot
        self.cov_log_det = self._cov_log_det()
        self.partial_bic = self.cov_log_det * 0.5 * self.count

    @classmethod
    def merge(cls, m1, m2):
        """
        Merge two models *m1* and *m2*. Compute the new mean (*mu*),
        covariance (*cov*) and PBIC *partial_bic*

        :param m1: a GaussFull object
        :param m2: a GaussFull object
        :return: a GaussFull object
        """
        m = GaussFull(m1.name, m1.dim)
        m.count = m1.count + m2.count
        m.stat1 = m1.stat1 + m2.stat1
        m.stat2 = m1.stat2 + m2.stat2
        m.mu = m.stat1 / m.count
        tmp = m.mu[:, np.newaxis]
        m.cov = m.stat2 / m.count - np.dot(tmp, tmp.T)
        m.cov_log_det = m._cov_log_det()
        m.partial_bic = m.cov_log_det * 0.5 * m.count
        return m

    @classmethod
    def merge_partial_bic(cls, m1, m2):
        """
        Merge statistic accumulators of two a GaussFull objects, compute
        the new log det.

        :param m1: a GaussFull object
        :param m2: a GaussFull object
        :return: the log det
        """
        try:
            count = m1.count + m2.count
            mu = ((m1.stat1 + m2.stat1) / count)[:, np.newaxis]
            cov = (m1.stat2 + m2.stat2) / count - np.dot(mu, mu.T)
            s, d = np.linalg.slogdet(cov)
            # cov_chol, lower = scipy.linalg.cho_factor(cov, overwrite_a=True, check_finite=False)
            # d = 2.0 * numpy.sum(numpy.log(numpy.diagonal(cov_chol)))
            d *= 0.5 * count
            return d
        except:
            logging.warning('Det problem set to NaN ', m1.name, m2.name)
            return np.nan

    @classmethod
    def cst_bic(cls, dim, alpha):
        """
        Compute the BIC constant:

            :math:`cst  = \\frac{1}{2} \\alpha \\left(d + \\frac{d(d+1)}{2}\\right)`

        where :math:`d`is the feature dimension (*dim*)
        and :math:`alpha` a threshold (*alpha*)

        :param dim: the feature dimension
        :param alpha: the threshold
        :return: the constant
        """
        return 0.5 * alpha * (dim + (0.5 * ((dim + 1) * dim)))







# -*- coding: utf-8 -*-
#
# This file is part of s4d.
#
# s4d is a python package for speaker diarization.
# Home page: http://www-lium.univ-lemans.fr/s4d/
#
# s4d is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# s4d is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with s4d.  If not, see <http://www.gnu.org/licenses/>.


"""
Copyright 2014-2021 Anthony Larcher
"""

import logging
import numpy
import shutil
import torch
import tqdm
import yaml

from collections import OrderedDict
from sidekit.nnet.sincnet import SincNet
from torch.utils.data import DataLoader

from .wavsets import SeqSet
from .wavsets import create_train_val_seqtoseq

__license__ = "LGPL"
__author__ = "Anthony Larcher, Martin Lebourdais, Meysam Shamsi"
__copyright__ = "Copyright 2015-2021 Anthony Larcher"
__maintainer__ = "Anthony Larcher"
__email__ = "anthony.larcher@univ-lemans.fr"
__status__ = "Production"
__docformat__ = 'reS'


logging.basicConfig(format='%(asctime)s %(message)s')


def save_checkpoint(state, is_best, filename='checkpoint.pth.tar', best_filename='model_best.pth.tar'):
    """
    :param state:
    :param is_best:
    :param filename:
    :param best_filename:
    :return:
    """
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, best_filename)


def init_weights(m):
    """

    :return:
    """
    if type(m) == torch.nn.Linear:
        torch.nn.init.xavier_uniform_(m.weight)
        m.bias.data.fill_(0.01)


def _unfold(final_output, output, shift, output_rate):
    """
    append the output of one batch sample with overlapp to the final output sequence without overlap
    The overlap part will be gotten sumation
    """

    # Case where the output is the target and not data
    if len(output.shape) < 2:
        output = numpy.array([output])

    len_seq = output.shape[1]
    sampleNumber = int(shift * output_rate)

    for i in range(len(output)):
        if len(final_output) == 0:
            final_output = numpy.array(output[i]).tolist()
        else:
            final_output = final_output[:-len_seq + sampleNumber] \
                           + (numpy.array(final_output[-len_seq + sampleNumber:]) + numpy.array(output[i])[
                                                                                    :-sampleNumber]).tolist() \
                           + numpy.array(output[i]).tolist()[-sampleNumber:]

    return final_output

def multi_label_combination(output_idx, output_target, output_data, shift, output_rate, mode="mean"):
    """

    :param output_idx:
    :param output_target:
    :param output_data:
    :param shift:
    :param output_rate:
    :param mode:
    :return:
    """
    win_shift = int(shift * output_rate)

    # Initialize the size of final_output
    final_output = numpy.zeros((win_shift * (len(output_data) - 1) + output_data[0].shape[0], output_data[0].shape[1]))

    final_target = numpy.zeros(win_shift * (len(output_data) - 1) + output_data[0].shape[0])
    overlaping_label_count = numpy.zeros(final_output.shape)

    win_len = output_data[0].shape[0]
    tmp = numpy.ones(output_data[0].shape)

    # Loop on the overlaping windows
    for idx, tmp_t, tmp_d in zip(output_idx, output_target, output_data):
        start_idx = win_shift * idx
        stop_idx = start_idx + win_len

        overlaping_label_count[start_idx: stop_idx, :] += tmp
        final_output[start_idx: stop_idx, :] += tmp_d
        final_target[start_idx: stop_idx] += tmp_t

    # Divide by the number of overlapping values
    final_output /= overlaping_label_count
    final_target /= overlaping_label_count[:, 0].squeeze()

    return final_output, final_target


def _output2sns_byoffset(model_out, output_offset):
    """
    Asign the output of model with 2 values to calss label based on index
    :param output_offset: if is None the calss labels will be maximum of values,
    otherwise it return thethe index of maximum of class1 and (class2 + output_offset)
    """

    if output_offset == None:
        return numpy.argmax(model_out, axis=1)
    else:
        seq = numpy.zeros(len(model_out), dtype=int)
        for i in range(len(model_out)):
            if model_out[i, 1] + output_offset > model_out[i, 0]:
                seq[i] = 1
            else:
                seq[i] = 0
        return seq


class BLSTM(torch.nn.Module):
    """
    Bi LSTM model used for voice activity detection, speaker turn detection, overlap detection and resegmentation
    """
    def __init__(self,
                 input_size,
                 blstm_sizes,
                 num_layers):
        """

        :param input_size:
        :param blstm_sizes:
        """
        super(BLSTM, self).__init__()
        self.input_size = input_size
        self.blstm_sizes = blstm_sizes
        self.output_size = blstm_sizes * 2

        self.blstm_layers = torch.nn.LSTM(input_size,
                                          blstm_sizes,
                                          bidirectional=True,
                                          batch_first=True,
                                          num_layers=num_layers)

    def forward(self, inputs):
        """

        :param inputs:
        :return:
        """
        output, h = self.blstm_layers(inputs)
        return output

    def output_size(self):
        """

        :return:
        """
        return self.output_size


class SeqToSeq(torch.nn.Module):
    """
    Model used for voice activity detection or speaker turn detection
    This model can include a pre-processor to input raw waveform,
    a BLSTM module to process the sequence-to-sequence
    and other linear of convolutional layers
    """
    def __init__(self,
                 model_archi):

        super(SeqToSeq, self).__init__()

        # Load Yaml configuration
        with open(model_archi, 'r') as fh:
            cfg = yaml.load(fh, Loader=yaml.FullLoader)

        self.loss = cfg["loss"]
        self.feature_size = None

        """
        Prepare Preprocessor
        """
        self.preprocessor = None
        if "preprocessor" in cfg:
            if cfg['preprocessor']["type"] == "sincnet":
                self.preprocessor = SincNet(
                    waveform_normalize=cfg['preprocessor']["waveform_normalize"],
                    sample_rate=cfg['preprocessor']["sample_rate"],
                    min_low_hz=cfg['preprocessor']["min_low_hz"],
                    min_band_hz=cfg['preprocessor']["min_band_hz"],
                    out_channels=cfg['preprocessor']["out_channels"],
                    kernel_size=cfg['preprocessor']["kernel_size"],
                    stride=cfg['preprocessor']["stride"],
                    max_pool=cfg['preprocessor']["max_pool"],
                    instance_normalize=cfg['preprocessor']["instance_normalize"],
                    activation=cfg['preprocessor']["activation"],
                    dropout=cfg['preprocessor']["dropout"]
                )
                self.feature_size = self.preprocessor.dimension

        """
        Prepare sequence to sequence  network
        """
        # Get Feature size
        if self.feature_size is None:
            self.feature_size = cfg["feature_size"]

        input_size = self.feature_size
        sequence_to_sequence_layers = []
        for k in cfg["sequence_to_sequence"].keys():
            if k.startswith("blstm"):
                sequence_to_sequence_layers.append((k, BLSTM(input_size=input_size,
                                                             blstm_sizes=cfg["sequence_to_sequence"][k]["output_size"],
                                                             num_layers=cfg["sequence_to_sequence"][k]["num_layers"])))
                input_size = cfg["sequence_to_sequence"][k]["output_size"] * 2

        self.sequence_to_sequence = torch.nn.Sequential(OrderedDict(sequence_to_sequence_layers))

        """
        Prepare post-processing network
        """
        # Create sequential object for the second part of the network
        self.post_processing_activation = torch.nn.Tanh()
        post_processing_layers = []
        for k in cfg["post_processing"].keys():

            if k.startswith("lin"):
                post_processing_layers.append((k, torch.nn.Linear(input_size,
                                                                  cfg["post_processing"][k]["output"])))
                input_size = cfg["post_processing"][k]["output"]

            elif k.startswith("activation"):
                post_processing_layers.append((k, self.post_processing_activation))

            elif k.startswith('batch_norm'):
                post_processing_layers.append((k, torch.nn.BatchNorm1d(input_size)))

            elif k.startswith('dropout'):
                post_processing_layers.append((k, torch.nn.Dropout(p=cfg["post_processing"][k])))

        self.post_processing = torch.nn.Sequential(OrderedDict(post_processing_layers))
        self.post_processing.apply(init_weights)

    def forward(self, inputs):
        """

        :param inputs:
        :return:
        """
        if self.preprocessor is not None:
            x = self.preprocessor(inputs)
            x = x.permute(0, 2, 1)
        else:
            x = inputs
        x = self.sequence_to_sequence(x)
        x = self.post_processing(x)
        return x

    def predict(self,
                batch_size,
                validation_loader,
                device,
                shift, # former step
                output_rate,
                th_in, #output_offset=None,
                th_out,
                only_lab_generation=False):
        """
            A MODIFIER POU NE PRENDRE QUE LE NOM DU FICHIER WAV ET CRÉER LE DATZA LOADER À L'INTERIEUR

        :param model:
        :param validation_loader:
        :param device:
        :param only_lab_generation: only generating lab file for each wav file without evaluation
        :return:
        """

        recall = 0.0
        precision = 0.0
        accuracy = 0.0
        loss = 0.0

        output_data = []
        output_target = []
        output_idx = []

        sm = torch.nn.Softmax(dim=2)
        with torch.no_grad():

            for batch_idx, (win_idx, data, target) in tqdm.tqdm(enumerate(validation_loader)):
                target = target.squeeze().cpu().numpy()
                output = sm(self.forward(data.to(device))).cpu().numpy()

                for ii in range(output.shape[0]):
                    output_data.append(output[ii])
                    output_target.append(target[ii])
                    output_idx.append(int(win_idx[ii]))

        # Unfold outputs by averaging sliding windows
        final_output, final_target = multi_label_combination(output_idx,
                                                             output_target,
                                                             output_data,
                                                             shift,
                                                             output_rate,
                                                             mode="mean")

        vad = numpy.zeros(final_output.shape[0], dtype='bool')
        speech = False
        ii = 0
        while ii < final_output.shape[0]:
            if final_output[ii, 1] > th_in and not speech:
                speech = True
            elif final_output[ii, 1] < th_out and speech:
                speech = False
            vad[ii] = speech
            ii += 1

        #if not only_lab_generation:
        #    final_target = _unfold(final_target, target.cpu().numpy(), shift)

        #sad_seq = output2sns_byoffset(numpy.array(final_output), threshold)

        #if not only_lab_generation:
        #    recall, precision, accuracy, tp, fp, tn, fn = calc_recall(sad_seq, (numpy.array(final_target) != 0),
        #                                                              uemfile_dir=uemfile_dir)

        #    if precision != 0 or recall != 0:
        #        f_measure = 2 * (precision) * (recall) / (precision + recall)
        #        logging.critical(
        #            'Validation: [{}/{} ({:.0f}%)]  Accuracy: {:.3f}  ' \
        #            'Recall: {:.3f}  Precision: {:.3f} ' \
        #            'F-Measure: {:.3f}'.format(batch_idx + 1,
        #                                       validation_loader.__len__(),
        #                                       100. * batch_idx / validation_loader.__len__(),
        #                                       100.0 * accuracy,
        #                                       100.0 * recall,
        #                                       100.0 * precision,
        #                                       f_measure)
        #        )
        #    return 100.0 * accuracy, 100.0 * recall, 100.0 * precision, sad_seq, tp, fp, tn, fn
        #else:
        #    return 0, 0, 0, sad_seq, 0, 0, 0, 0
        return vad

def seqTrain(dataset_yaml,
             model_yaml,
             epochs=100,
             lr=0.0001,
             patience=10,
             model_name=None,
             tmp_model_name=None,
             best_model_name=None,
             multi_gpu=True,
             opt='sgd',
             log_interval=10,
             num_thread=10
             ):
    """

    :param data_dir:
    :param mode:
    :param duration:
    :param seg_shift:
    :param filter_type:
    :param collar_duration:
    :param framerate:
    :param epochs:
    :param lr:
    :param loss:
    :param patience:
    :param tmp_model_name:
    :param best_model_name:
    :param multi_gpu:
    :param opt:
    :return:
    """
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # Start from scratch
    if model_name is None:
       model = SeqToSeq(model_yaml)
    # If we start from an existing model
    else:
        # Load the model
        logging.critical(f"*** Load model from = {model_name}")
        checkpoint = torch.load(model_name, map_location='cpu')
        model = SeqToSeq(model_yaml)
        model.load_state_dict(checkpoint['model_state_dict'])

    if torch.cuda.device_count() > 1 and multi_gpu:
        print("Let's use", torch.cuda.device_count(), "GPUs!")
        model = torch.nn.DataParallel(model)
    else:
        print("Train on a single GPU")

    model.to(device)

    with open(dataset_yaml, "r") as fh:
        dataset_params = yaml.load(fh, Loader=yaml.FullLoader)
    """
    Create two dataloaders for training and evaluation
    """
    training_set, validation_set = create_train_val_seqtoseq(dataset_yaml)

    training_loader = DataLoader(training_set,
                                 batch_size=dataset_params["batch_size"],
                                 shuffle=True,
                                 drop_last=True,
                                 pin_memory=True,
                                 num_workers=num_thread)

    validation_loader = DataLoader(validation_set,
                                   batch_size=dataset_params["batch_size"],
                                   drop_last=True,
                                   pin_memory=True,
                                   num_workers=num_thread)

    """
    Set the training options
    """
    if opt == 'sgd':
        _optimizer = torch.optim.SGD
        _options = {'lr': lr, 'momentum': 0.9}
    elif opt == 'adam':
        _optimizer = torch.optim.Adam
        _options = {'lr': lr}
    elif opt == 'rmsprop':
        _optimizer = torch.optim.RMSprop
        _options = {'lr': lr}

    params = [
        {
            'params': [
                param for name, param in model.named_parameters() if 'bn' not in name
            ]
        },
        {
            'params': [
                param for name, param in model.named_parameters() if 'bn' in name
            ],
            'weight_decay': 0
        },
    ]

    optimizer = _optimizer([{'params': model.parameters()},], **_options)

    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', verbose=True)

    best_accuracy = 0.0
    best_accuracy_epoch = 1
    curr_patience = patience

    for epoch in range(1, epochs + 1):
        # Process one epoch and return the current model
        if curr_patience == 0:
            print(f"Stopping at epoch {epoch} for cause of patience")
            break
        model = train_epoch(model,
                            epoch,
                            training_loader,
                            optimizer,
                            log_interval,
                            device=device)

        # Cross validation here
        accuracy, val_loss = cross_validation(model, validation_loader, device=device)
        logging.critical("*** Cross validation accuracy = {} %".format(accuracy))

        # Decrease learning rate according to the scheduler policy
        scheduler.step(val_loss)
        print(f"Learning rate is {optimizer.param_groups[0]['lr']}")

        # remember best accuracy and save checkpoint
        is_best = accuracy > best_accuracy
        best_accuracy = max(accuracy, best_accuracy)

        if type(model) is SeqToSeq:
            save_checkpoint({
                'epoch': epoch,
                'model_state_dict': model.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
                'accuracy': best_accuracy,
                'scheduler': scheduler
            }, is_best, filename=tmp_model_name + ".pt", best_filename=best_model_name + '.pt')
        else:
            save_checkpoint({
                'epoch': epoch,
                'model_state_dict': model.module.state_dict(),
                'optimizer_state_dict': optimizer.state_dict(),
                'accuracy': best_accuracy,
                'scheduler': scheduler
            }, is_best, filename=tmp_model_name + ".pt", best_filename=best_model_name + '.pt')

        if is_best:
            best_accuracy_epoch = epoch
            curr_patience = patience
        else:
            curr_patience -= 1

    logging.critical(f"Best accuracy {best_accuracy * 100.} obtained at epoch {best_accuracy_epoch}")


def train_epoch(model, epoch, training_loader, optimizer, log_interval, device):
    """

    :param model:
    :param epoch:
    :param training_loader:
    :param optimizer:
    :param log_interval:
    :param device:
    :param clipping:
    :return:
    """
    model.to(device)
    model.train()
    criterion = torch.nn.CrossEntropyLoss(reduction='mean', weight=torch.FloatTensor([0.9,0.1]).to(device))

    recall = 0.0
    precision = 0.0
    accuracy = 0.0

    for batch_idx, (data, target) in enumerate(training_loader):
        target = target.squeeze()
        optimizer.zero_grad()
        output = model(data.to(device))
        output = output.permute(1, 2, 0)
        target = target.permute(1, 0)

        loss = criterion(output, target.to(device))
        loss.backward(retain_graph=True)
        optimizer.step()

        rc, pr, acc = calc_recall(output.data, target, device)
        recall += rc.item()
        precision += pr.item()
        accuracy += acc.item()

        if batch_idx % log_interval == 0:
            batch_size = target.shape[0]
            if precision!=0 or recall!=0:
                f_measure = 2 * (precision / ((batch_idx + 1))) * (recall / ((batch_idx+1))) /\
                            ((precision / ((batch_idx + 1) ))+(recall / ((batch_idx + 1))))
                logging.critical(
                        'Train Epoch: {} [{}/{} ({:.0f}%)]  Loss: {:.6f}  Accuracy: {:.3f}  '\
                        'Recall: {:.3f}  Precision: {:.3f}  '\
                        'F-Measure: {:.3f}'.format(epoch,
                                                  batch_idx + 1,
                                                  training_loader.__len__(),
                                                  100. * batch_idx / training_loader.__len__(), loss.item(),
                                                  100.0 * accuracy / ((batch_idx + 1)),
                                                  100.0 * recall/ ((batch_idx+1)),
                                                  100.0 * precision / ((batch_idx+1)),
                                                  f_measure)
                )
            else:
                print(f"precision = {precision} and recall = {recall}")

    return model


def cross_validation(model, validation_loader, device):
    """

    :param model:
    :param validation_loader:
    :param device:
    :return:
    """
    model.eval()

    recall = 0.0
    precision = 0.0
    accuracy = 0.0
    loss = 0.0

    criterion = torch.nn.CrossEntropyLoss()
    with torch.no_grad():
        for batch_idx, (data, target) in enumerate(validation_loader):
            target = target.squeeze()
            output = model(data.to(device))
            output = output.permute(1, 2, 0)
            target = target.permute(1, 0)

            loss += criterion(output, target.to(device))

            rc, pr, acc = calc_recall(output.data, target, device)
            recall += rc.item()
            precision += pr.item()
            accuracy += acc.item()

    batch_size = target.shape[0]
    if precision != 0 or recall != 0:
        f_measure = 2 * (precision / ((batch_idx + 1))) * (recall / ((batch_idx + 1))) / \
                    ((precision / ((batch_idx + 1))) + (recall / ((batch_idx + 1))))
        logging.critical(
            'Validation: [{}/{} ({:.0f}%)]  Loss: {:.6f}  Accuracy: {:.3f}  ' \
            'Recall: {:.3f}  Precision: {:.3f}  '\
            'F-Measure: {:.3f}'.format(batch_idx + 1,
                                      validation_loader.__len__(),
                                      100. * batch_idx / validation_loader.__len__(),
                                      loss.item() / (batch_idx + 1),
                                      100.0 * accuracy / ((batch_idx + 1)),
                                      100.0 * recall / ((batch_idx + 1)),
                                      100.0 * precision / ((batch_idx + 1)),
                                      f_measure)
        )

    return 100.0 * accuracy / ((batch_idx + 1)), loss/(batch_idx + 1)


def calc_recall(output,target,device):
    """

    :param output:
    :param target:
    :param device:
    :return:
    """
    y_trueb = target.to(device)
    y_predb = output
    rc = 0.0
    pr = 0.0
    acc= 0.0
    for b in range(y_trueb.shape[-1]):
        y_true = y_trueb[:,b]
        y_pred = y_predb[:,:,b]
        assert y_true.ndim == 1
        assert y_pred.ndim == 1 or y_pred.ndim == 2
        if y_pred.ndim == 2:
            y_pred = y_pred.argmax(dim=1)

        tp = (y_true * y_pred).sum().to(torch.float32)
        tn = ((1 - y_true) * (1 - y_pred)).sum().to(torch.float32)
        fp = ((1 - y_true) * y_pred).sum().to(torch.float32)
        fn = (y_true * (1 - y_pred)).sum().to(torch.float32)
        epsilon = 1e-7

        pr+= tp / (tp + fp + epsilon)
        rc+= tp / (tp + fn + epsilon)
        a=(tp+tn)/(tp+fp+tn+fn+epsilon)

        acc+=(tp+tn)/(tp+fp+tn+fn+epsilon)
    rc/=len(y_trueb[0])
    pr/=len(y_trueb[0])
    acc/=len(y_trueb[0])

    return rc,pr,acc






#!/usr/bin/env python
# -*- coding: utf-8 -*-
 
from distutils.core import setup

# patch distutils if it can't cope with the "classifiers" or
# "download_url" keywords
from sys import version
if version < '2.2.3':
    from distutils.dist import DistributionMetadata
    DistributionMetadata.classifiers = None
    DistributionMetadata.download_url = None

def get_s4d_version():
    with open('s4d/__init__.py') as f:
        s = [l.rstrip() for l in f.readlines()]
        version = None
        for l in s:
            if '__version__' in l:
                version = l.split('=')[-1]
        if version is None:
            raise RuntimeError('Can not detect version from s4d/__init__.py')
        return eval(version)

setup(
    name='s4d',
    version=get_s4d_version(),
    author='Sylvain MEIGNIER',
    author_email='s4d@univ-lemans.fr',
    packages=['s4d', 's4d.alien', 's4d.clustering', 's4d.gui', 's4d.nnet'],
    url='https://projets-lium.univ-lemans.fr/s4d/',
    download_url='http://pypi.python.org/pypi/s4d/',
    license='LGPL',
    platforms=['Linux, Windows', 'MacOS'],
    description='S4D: SIDEKIT for Diarization',
    long_description=open('README.txt').read(),
    install_requires=[
        "numpy==1.19.0",
        "pyparsing >= 2.4.7",
        "python-dateutil >= 2.8.1",
        "scipy>=1.6.0",
        "six>=1.15.0",
        "matplotlib>=3.3.2",
        "soundfile>= 0.10.3",
        "torch >= 1.7.1",
        "torchvision>=0.8.2",
        "tqdm>=4.55.1",
        "pyyaml>=5.4.1",
        "h5py>=2.10.0",
        "pandas>=1.2.1",
        "scikit-learn>=0.24.1",
        "bottleneck>=1.3.1",
        "setuptools>=38.5.2",
        "sidekit>=1.3.8.3",
        "sortedcontainers>=1.5.9"
    ],
    package_data={'s4d': ['docs/*']},
    classifiers=[        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering",
        ]
)





# Hidden Markov Models
#
# Author: Ron Weiss <ronweiss@gmail.com>


import numpy as np
import string

from sklearn.base import BaseEstimator
from sklearn import cluster
from ..alien.mixture import GMM, sample_gaussian, _validate_covars
from ..alien.mixture import logsum, lmvnpdf, normalize, _distribute_covar_matrix_to_match_cvtype

ZEROLOGPROB = -1e200


class _BaseHMM(BaseEstimator):
    """Hidden Markov Model base class.

    Representation of a hidden Markov model probability distribution.
    This class allows for easy evaluation of, sampling from, and
    maximum-likelihood estimation of the parameters of a HMM.

    See the instance documentation for details specific to a
    particular object.

    Attributes
    ----------
    n_states : int (read-only)
        Number of states in the model.
    transmat : array, shape (`n_states`, `n_states`)
        Matrix of transition probabilities between states.
    startprob : array, shape ('n_states`,)
        Initial state occupation distribution.

    Methods
    -------
    eval(X)
        Compute the log likelihood of `X` under the HMM.
    decode(X)
        Find most likely state sequence for each point in `X` using the
        Viterbi algorithm.
    rvs(n=1)
        Generate `n` samples from the HMM.
    fit(X)
        Estimate HMM parameters from `X`.
    predict(X)
        Like decode, find most likely state sequence corresponding to `X`.
    score(X)
        Compute the log likelihood of `X` under the model.

    See Also
    --------
    GMM : Gaussian mixture model
    """

    # This class implements the public interface to all HMMs that
    # derive from it, including all of the machinery for the
    # forward-backward and Viterbi algorithms.  Subclasses need only
    # implement _generate_sample_from_state(), _compute_log_likelihood(),
    # _init(), _initialize_sufficient_statistics(),
    # _accumulate_sufficient_statistics(), and _do_mstep(), all of
    # which depend on the specific emission distribution.
    #
    # Subclasses will probably also want to implement properties for
    # the emission distribution parameters to expose them publically.

    def __init__(self, n_states=1, startprob=None, transmat=None,
                 startprob_prior=None, transmat_prior=None):
        self._n_states = n_states

        if startprob is None:
            startprob = np.tile(1.0 / n_states, n_states)
        self.startprob = startprob

        if startprob_prior is None:
            startprob_prior = 1.0
        self.startprob_prior = startprob_prior

        if transmat is None:
            transmat = np.tile(1.0 / n_states, (n_states, n_states))
        self.transmat = transmat

        if transmat_prior is None:
            transmat_prior = 1.0
        self.transmat_prior = transmat_prior

    def eval(self, obs, maxrank=None, beamlogprob=-np.Inf):
        """Compute the log probability under the model and compute posteriors

        Implements rank and beam pruning in the forward-backward
        algorithm to speed up inference in large models.

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            Sequence of n_features-dimensional data points.  Each row
            corresponds to a single point in the sequence.
        maxrank : int
            Maximum rank to evaluate for rank pruning.  If not None,
            only consider the top `maxrank` states in the inner
            sum of the forward algorithm recursion.  Defaults to None
            (no rank pruning).  See The HTK Book for more details.
        beamlogprob : float
            Width of the beam-pruning beam in log-probability units.
            Defaults to -numpy.Inf (no beam pruning).  See The HTK
            Book for more details.

        Returns
        -------
        logprob : array_like, shape (n,)
            Log probabilities of the sequence `observation`
        posteriors: array_like, shape (n, n_states)
            Posterior probabilities of each state for each
            observation

        See Also
        --------
        score : Compute the log probability under the model
        decode : Find most likely state sequence corresponding to a `observation`
        """
        obs = np.asanyarray(obs)
        framelogprob = self._compute_log_likelihood(obs)
        logprob, fwdlattice = self._do_forward_pass(framelogprob, maxrank,
                                                    beamlogprob)
        bwdlattice = self._do_backward_pass(framelogprob, fwdlattice, maxrank,
                                            beamlogprob)
        gamma = fwdlattice + bwdlattice
        # gamma is guaranteed to be correctly normalized by logprob at
        # all frames, unless we do approximate inference using pruning.
        # So, we will normalize each frame explicitly in case we
        # pruned too aggressively.
        posteriors = np.exp(gamma.T - logsum(gamma, axis=1)).T
        return logprob, posteriors

    def score(self, obs, maxrank=None, beamlogprob=-np.Inf):
        """Compute the log probability under the model.

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            Sequence of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        maxrank : int
            Maximum rank to evaluate for rank pruning.  If not None,
            only consider the top `maxrank` states in the inner
            sum of the forward algorithm recursion.  Defaults to None
            (no rank pruning).  See The HTK Book for more details.
        beamlogprob : float
            Width of the beam-pruning beam in log-probability units.
            Defaults to -numpy.Inf (no beam pruning).  See The HTK
            Book for more details.

        Returns
        -------
        logprob : array_like, shape (n,)
            Log probabilities of each data point in `observation`

        See Also
        --------
        eval : Compute the log probability under the model and posteriors
        decode : Find most likely state sequence corresponding to a `observation`
        """
        obs = np.asanyarray(obs)
        framelogprob = self._compute_log_likelihood(obs)
        logprob, fwdlattice = self._do_forward_pass(framelogprob, maxrank,
                                                    beamlogprob)
        return logprob

    def decode(self, obs, maxrank=None, beamlogprob=-np.Inf):
        """Find most likely state sequence corresponding to `observation`.

        Uses the Viterbi algorithm.

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        maxrank : int
            Maximum rank to evaluate for rank pruning.  If not None,
            only consider the top `maxrank` states in the inner
            sum of the forward algorithm recursion.  Defaults to None
            (no rank pruning).  See The HTK Book for more details.
        beamlogprob : float
            Width of the beam-pruning beam in log-probability units.
            Defaults to -numpy.Inf (no beam pruning).  See The HTK
            Book for more details.

        Returns
        -------
        viterbi_logprob : float
            Log probability of the maximum likelihood path through the HMM
        states : array_like, shape (n,)
            Index of the most likely states for each observation

        See Also
        --------
        eval : Compute the log probability under the model and posteriors
        score : Compute the log probability under the model
        """
        obs = np.asanyarray(obs)
        framelogprob = self._compute_log_likelihood(obs)
        logprob, state_sequence = self._do_viterbi_pass(framelogprob, maxrank,
                                                        beamlogprob)
        return logprob, state_sequence

    def predict(self, obs, **kwargs):
        """Find most likely state sequence corresponding to `observation`.

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.
        maxrank : int
            Maximum rank to evaluate for rank pruning.  If not None,
            only consider the top `maxrank` states in the inner
            sum of the forward algorithm recursion.  Defaults to None
            (no rank pruning).  See The HTK Book for more details.
        beamlogprob : float
            Width of the beam-pruning beam in log-probability units.
            Defaults to -numpy.Inf (no beam pruning).  See The HTK
            Book for more details.

        Returns
        -------
        states : array_like, shape (n,)
            Index of the most likely states for each observation
        """
        logprob, state_sequence = self.decode(obs, **kwargs)
        return state_sequence

    def predict_proba(self, obs, **kwargs):
        """Compute the posterior probability for each state in the model

        Parameters
        ----------
        obs : array_like, shape (n, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        See eval() for a list of accepted keyword arguments.

        Returns
        -------
        T : array-like, shape (n, n_states)
            Returns the probability of the sample for each state in the model.
        """
        logprob, posteriors = self.eval(obs, **kwargs)
        return posteriors

    def rvs(self, n=1):
        """Generate random samples from the model.

        Parameters
        ----------
        n : int
            Number of samples to generate.

        Returns
        -------
        observation : array_like, length `n`
            List of samples
        """

        startprob_pdf = self.startprob
        startprob_cdf = np.cumsum(startprob_pdf)
        transmat_pdf = self.transmat
        transmat_cdf = np.cumsum(transmat_pdf, 1)

        # Initial state.
        rand = np.random.rand()
        currstate = (startprob_cdf > rand).argmax()
        obs = [self._generate_sample_from_state(currstate)]

        for x in range(n - 1):
            rand = np.random.rand()
            currstate = (transmat_cdf[currstate] > rand).argmax()
            obs.append(self._generate_sample_from_state(currstate))

        return np.array(obs)

    def fit(self, obs, n_iter=10, thresh=1e-2, params=string.ascii_letters,
            init_params=string.ascii_letters,
            maxrank=None, beamlogprob=-np.Inf, **kwargs):
        """Estimate model parameters.

        An initialization step is performed before entering the EM
        algorithm. If you want to avoid this step, set the keyword
        argument init_params to the empty string ''. Likewise, if you
        would like just to do an initialization, call this method with
        n_iter=0.

        Parameters
        ----------
        obs : list
            List of array-like observation sequences (shape (n_i, n_features)).
        n_iter : int, optional
            Number of iterations to perform.
        thresh : float, optional
            Convergence threshold.
        params : string, optional
            Controls which parameters are updated in the training
            process.  Can contain any combination of 's' for startprob,
            't' for transmat, 'm' for means, and 'c' for covars, etc.
            Defaults to all parameters.
        init_params : string, optional
            Controls which parameters are initialized prior to
            training.  Can contain any combination of 's' for
            startprob, 't' for transmat, 'm' for means, and 'c' for
            covars, etc.  Defaults to all parameters.
        maxrank : int, optional
            Maximum rank to evaluate for rank pruning.  If not None,
            only consider the top `maxrank` states in the inner
            sum of the forward algorithm recursion.  Defaults to None
            (no rank pruning).  See "The HTK Book" for more details.
        beamlogprob : float, optional
            Width of the beam-pruning beam in log-probability units.
            Defaults to -numpy.Inf (no beam pruning).  See "The HTK
            Book" for more details.

        Notes
        -----
        In general, `logprob` should be non-decreasing unless
        aggressive pruning is used.  Decreasing `logprob` is generally
        a sign of overfitting (e.g. a covariance parameter getting too
        small).  You can fix this by getting more training data, or
        decreasing `covars_prior`.
        """
        self._init(obs, init_params)

        logprob = []
        for i in range(n_iter):
            # Expectation step
            stats = self._initialize_sufficient_statistics()
            curr_logprob = 0
            for seq in obs:
                framelogprob = self._compute_log_likelihood(seq)
                lpr, fwdlattice = self._do_forward_pass(framelogprob, maxrank,
                                                       beamlogprob)
                bwdlattice = self._do_backward_pass(framelogprob, fwdlattice,
                                                   maxrank, beamlogprob)
                gamma = fwdlattice + bwdlattice
                posteriors = np.exp(gamma.T - logsum(gamma, axis=1)).T
                curr_logprob += lpr
                self._accumulate_sufficient_statistics(
                    stats, seq, framelogprob, posteriors, fwdlattice,
                    bwdlattice, params)
            logprob.append(curr_logprob)

            # Check for convergence.
            if i > 0 and abs(logprob[-1] - logprob[-2]) < thresh:
                break

            # Maximization step
            self._do_mstep(stats, params, **kwargs)

        return self

    @property
    def n_states(self):
        """Number of states in the model."""
        return self._n_states

    def _get_startprob(self):
        """Mixing startprob for each state."""
        return np.exp(self._log_startprob)

    def _set_startprob(self, startprob):
        if len(startprob) != self._n_states:
            raise ValueError('startprob must have length n_states')
        if not np.allclose(np.sum(startprob), 1.0):
            raise ValueError('startprob must sum to 1.0')

        self._log_startprob = np.log(np.asanyarray(startprob).copy())

    startprob = property(_get_startprob, _set_startprob)

    def _get_transmat(self):
        """Matrix of transition probabilities."""
        return np.exp(self._log_transmat)

    def _set_transmat(self, transmat):
        if np.asanyarray(transmat).shape != (self._n_states, self._n_states):
            raise ValueError('transmat must have shape (n_states, n_states)')
        if not np.all(np.allclose(np.sum(transmat, axis=1), 1.0)):
            raise ValueError('Rows of transmat must sum to 1.0')

        self._log_transmat = np.log(np.asanyarray(transmat).copy())
        underflow_idx = np.isnan(self._log_transmat)
        self._log_transmat[underflow_idx] = -np.Inf

    transmat = property(_get_transmat, _set_transmat)

    def _do_viterbi_pass(self, framelogprob, maxrank=None,
                         beamlogprob=-np.Inf):
        nobs = len(framelogprob)
        lattice = np.zeros((nobs, self._n_states))
        traceback = np.zeros((nobs, self._n_states), dtype=np.int)

        lattice[0] = self._log_startprob + framelogprob[0]
        for n in range(1, nobs):
            idx = self._prune_states(lattice[n - 1], maxrank, beamlogprob)
            pr = self._log_transmat[idx].T + lattice[n - 1, idx]
            lattice[n] = np.max(pr, axis=1) + framelogprob[n]
            traceback[n] = np.argmax(pr, axis=1)
        lattice[lattice <= ZEROLOGPROB] = -np.Inf

        # Do traceback.
        reverse_state_sequence = []
        s = lattice[-1].argmax()
        logprob = lattice[-1, s]
        for frame in reversed(traceback):
            reverse_state_sequence.append(s)
            s = frame[s]

        reverse_state_sequence.reverse()
        return logprob, np.array(reverse_state_sequence)

    def _do_forward_pass(self, framelogprob, maxrank=None,
                         beamlogprob=-np.Inf):
        nobs = len(framelogprob)
        fwdlattice = np.zeros((nobs, self._n_states))

        fwdlattice[0] = self._log_startprob + framelogprob[0]
        for n in range(1, nobs):
            idx = self._prune_states(fwdlattice[n - 1], maxrank, beamlogprob)
            fwdlattice[n] = (logsum(self._log_transmat[idx].T
                                    + fwdlattice[n - 1, idx], axis=1)
                             + framelogprob[n])
        fwdlattice[fwdlattice <= ZEROLOGPROB] = -np.Inf

        return logsum(fwdlattice[-1]), fwdlattice

    def _do_backward_pass(self, framelogprob, fwdlattice, maxrank=None,
                          beamlogprob=-np.Inf):
        nobs = len(framelogprob)
        bwdlattice = np.zeros((nobs, self._n_states))

        for n in range(nobs - 1, 0, -1):
            # Do HTK style pruning (p. 137 of HTK Book version 3.4).
            # Don't bother computing backward probability if
            # fwdlattice * bwdlattice is more than a certain distance
            # from the total log likelihood.
            idx = self._prune_states(bwdlattice[n] + fwdlattice[n], None,
                                     -50)
                                     #beamlogprob)
                                     #-numpy.Inf)
            bwdlattice[n - 1] = logsum(self._log_transmat[:, idx] +
                                       bwdlattice[n, idx] +
                                       framelogprob[n, idx],
                                       axis=1)
        bwdlattice[bwdlattice <= ZEROLOGPROB] = -np.Inf

        return bwdlattice

    def _prune_states(self, lattice_frame, maxrank, beamlogprob):
        """ Returns indices of the active states in `lattice_frame`
        after rank and beam pruning.
        """
        # Beam pruning
        threshlogprob = logsum(lattice_frame) + beamlogprob

        # Rank pruning
        if maxrank:
            # How big should our rank pruning histogram be?
            nbins = 3 * len(lattice_frame)

            lattice_min = lattice_frame[lattice_frame > ZEROLOGPROB].min() - 1
            hst, cdf = np.histogram(lattice_frame, bins=nbins,
                                    range=(lattice_min, lattice_frame.max()))

            # Want to look at the high ranks.
            hst = hst[::-1].cumsum()
            cdf = cdf[::-1]

            rankthresh = cdf[hst >= min(maxrank, self._n_states)].max()

            # Only change the threshold if it is stricter than the beam
            # threshold.
            threshlogprob = max(threshlogprob, rankthresh)

        # Which states are active?
        state_idx, = np.nonzero(lattice_frame >= threshlogprob)
        return state_idx

    def _compute_log_likelihood(self, obs):
        pass

    def _generate_sample_from_state(self, state):
        pass

    def _init(self, obs, params):
        if 's' in params:
            self.startprob[:] = 1.0 / self._n_states
        if 't' in params:
            self.transmat[:] = 1.0 / self._n_states

    # Methods used by self.fit()

    def _initialize_sufficient_statistics(self):
        stats = {'nobs': 0,
                 'start': np.zeros(self._n_states),
                 'trans': np.zeros((self._n_states, self._n_states))}
        return stats

    def _accumulate_sufficient_statistics(self, stats, seq, framelogprob,
                                          posteriors, fwdlattice, bwdlattice,
                                          params):
        stats['nobs'] += 1
        if 's' in params:
            stats['start'] += posteriors[0]
        if 't' in params:
            for t in range(len(framelogprob)):
                zeta = (fwdlattice[t - 1][:, np.newaxis] + self._log_transmat
                        + framelogprob[t] + bwdlattice[t])
                stats['trans'] += np.exp(zeta - logsum(zeta))

    def _do_mstep(self, stats, params, **kwargs):
        # Based on Huang, Acero, Hon, "Spoken Language Processing",
        # p. 443 - 445
        if 's' in params:
            self.startprob = normalize(
                np.maximum(self.startprob_prior - 1.0 + stats['start'], 1e-20))
        if 't' in params:
            self.transmat = normalize(
                np.maximum(self.transmat_prior - 1.0 + stats['trans'], 1e-20),
                axis=1)


class GaussianHMM(_BaseHMM):
    """Hidden Markov Model with Gaussian emissions

    Representation of a hidden Markov model probability distribution.
    This class allows for easy evaluation of, sampling from, and
    maximum-likelihood estimation of the parameters of a HMM.

    Attributes
    ----------
    cvtype : string (read-only)
        String describing the type of covariance parameters used by
        the model.  Must be one of 'spherical', 'tied', 'diag', 'full'.
    n_features : int (read-only)
        Dimensionality of the Gaussian emissions.
    n_states : int (read-only)
        Number of states in the model.
    transmat : array, shape (`n_states`, `n_states`)
        Matrix of transition probabilities between states.
    startprob : array, shape ('n_states`,)
        Initial state occupation distribution.
    means : array, shape (`n_states`, `n_features`)
        Mean parameters for each state.
    covars : array
        Covariance parameters for each state.  The shape depends on
        `cvtype`:
            (`n_states`,)                   if 'spherical',
            (`n_features`, `n_features`)              if 'tied',
            (`n_states`, `n_features`)           if 'diag',
            (`n_states`, `n_features`, `n_features`)  if 'full'

    Methods
    -------
    eval(X)
        Compute the log likelihood of `X` under the HMM.
    decode(X)
        Find most likely state sequence for each point in `X` using the
        Viterbi algorithm.
    rvs(n=1)
        Generate `n` samples from the HMM.
    init(X)
        Initialize HMM parameters from `X`.
    fit(X)
        Estimate HMM parameters from `X` using the Baum-Welch algorithm.
    predict(X)
        Like decode, find most likely state sequence corresponding to `X`.
    score(X)
        Compute the log likelihood of `X` under the model.

    Examples
    --------
    >>> GaussianHMM(n_states=2)
    GaussianHMM(cvtype='diag', n_states=2, means_weight=0, startprob_prior=1.0,
          startprob=array([ 0.5,  0.5]),
          transmat=array([[ 0.5,  0.5],
           [ 0.5,  0.5]]),
          transmat_prior=1.0, means_prior=None, covars_weight=1,
          covars_prior=0.01)

    See Also
    --------
    GMM : Gaussian mixture model
    """

    def __init__(self, n_states=1, cvtype='diag', startprob=None,
                 transmat=None, startprob_prior=None, transmat_prior=None,
                 means_prior=None, means_weight=0,
                 covars_prior=1e-2, covars_weight=1):
        """Create a hidden Markov model with Gaussian emissions.

        Initializes parameters such that every state has zero mean and
        identity covariance.

        Parameters
        ----------
        n_states : int
            Number of states.
        cvtype : string
            String describing the type of covariance parameters to
            use.  Must be one of 'spherical', 'tied', 'diag', 'full'.
            Defaults to 'diag'.
        """
        super(GaussianHMM, self).__init__(n_states, startprob, transmat,
                                          startprob_prior=startprob_prior,
                                          transmat_prior=transmat_prior)

        self._cvtype = cvtype
        if not cvtype in ['spherical', 'tied', 'diag', 'full']:
            raise ValueError('bad cvtype')

        self.means_prior = means_prior
        self.means_weight = means_weight

        self.covars_prior = covars_prior
        self.covars_weight = covars_weight

    # Read-only properties.
    @property
    def cvtype(self):
        """Covariance type of the model.

        Must be one of 'spherical', 'tied', 'diag', 'full'.
        """
        return self._cvtype

    def _get_means(self):
        """Mean parameters for each state."""
        return self._means

    def _set_means(self, means):
        means = np.asanyarray(means)
        if hasattr(self, 'n_features') and \
               means.shape != (self._n_states, self.n_features):
            raise ValueError('means must have shape (n_states, n_features)')
        self._means = means.copy()
        self.n_features = self._means.shape[1]

    means = property(_get_means, _set_means)

    def _get_covars(self):
        """Return covars as a full matrix."""
        if self.cvtype == 'full':
            return self._covars
        elif self.cvtype == 'diag':
            return [np.diag(cov) for cov in self._covars]
        elif self.cvtype == 'tied':
            return [self._covars] * self._n_states
        elif self.cvtype == 'spherical':
            return [np.eye(self.n_features) * f for f in self._covars]

    def _set_covars(self, covars):
        covars = np.asanyarray(covars)
        _validate_covars(covars, self._cvtype, self._n_states, self.n_features)
        self._covars = covars.copy()

    covars = property(_get_covars, _set_covars)

    def _compute_log_likelihood(self, obs):
        return lmvnpdf(obs, self._means, self._covars, self._cvtype)

    def _generate_sample_from_state(self, state):
        if self._cvtype == 'tied':
            cv = self._covars
        else:
            cv = self._covars[state]
        return sample_gaussian(self._means[state], cv, self._cvtype)

    def _init(self, obs, params='stmc'):
        super(GaussianHMM, self)._init(obs, params=params)

        if (hasattr(self, 'n_features')
            and self.n_features != obs[0].shape[1]):
            raise ValueError('Unexpected number of dimensions, got %s but '
                             'expected %s' % (obs[0].shape[1],
                                              self.n_features))

        self.n_features = obs[0].shape[1]

        if 'm' in params:
            self._means = cluster.KMeans(
                k=self._n_states).fit(obs[0]).cluster_centers_
        if 'c' in params:
            cv = np.cov(obs[0].T)
            if not cv.shape:
                cv.shape = (1, 1)
            self._covars = _distribute_covar_matrix_to_match_cvtype(
                cv, self._cvtype, self._n_states)

    def _initialize_sufficient_statistics(self):
        stats = super(GaussianHMM, self)._initialize_sufficient_statistics()
        stats['post'] = np.zeros(self._n_states)
        stats['observation'] = np.zeros((self._n_states, self.n_features))
        stats['observation**2'] = np.zeros((self._n_states, self.n_features))
        stats['observation*observation.T'] = np.zeros((self._n_states, self.n_features,
                                       self.n_features))
        return stats

    def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                          posteriors, fwdlattice, bwdlattice,
                                          params):
        super(GaussianHMM, self)._accumulate_sufficient_statistics(
            stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice,
            params)

        if 'm' in params or 'c' in params:
            stats['post'] += posteriors.sum(axis=0)
            stats['observation'] += np.dot(posteriors.T, obs)

        if 'c' in params:
            if self._cvtype in ('spherical', 'diag'):
                stats['observation**2'] += np.dot(posteriors.T, obs ** 2)
            elif self._cvtype in ('tied', 'full'):
                for t, o in enumerate(obs):
                    obsobsT = np.outer(o, o)
                    for c in range(self._n_states):
                        stats['observation*observation.T'][c] += posteriors[t, c] * obsobsT

    def _do_mstep(self, stats, params, **kwargs):
        super(GaussianHMM, self)._do_mstep(stats, params)

        # Based on Huang, Acero, Hon, "Spoken Language Processing",
        # p. 443 - 445
        denom = stats['post'][:, np.newaxis]
        if 'm' in params:
            prior = self.means_prior
            weight = self.means_weight
            if prior is None:
                weight = 0
                prior = 0
            self._means = (weight * prior + stats['observation']) / (weight + denom)

        if 'c' in params:
            covars_prior = self.covars_prior
            covars_weight = self.covars_weight
            if covars_prior is None:
                covars_weight = 0
                covars_prior = 0

            means_prior = self.means_prior
            means_weight = self.means_weight
            if means_prior is None:
                means_weight = 0
                means_prior = 0
            meandiff = self._means - means_prior

            if self._cvtype in ('spherical', 'diag'):
                cv_num = (means_weight * (meandiff) ** 2
                          + stats['observation**2']
                          - 2 * self._means * stats['observation']
                          + self._means ** 2 * denom)
                cv_den = max(covars_weight - 1, 0) + denom
                if self._cvtype == 'spherical':
                    self._covars = (covars_prior / cv_den.mean(axis=1)
                                   + np.mean(cv_num / cv_den, axis=1))
                elif self._cvtype == 'diag':
                    self._covars = (covars_prior + cv_num) / cv_den
            elif self._cvtype in ('tied', 'full'):
                cvnum = np.empty((self._n_states, self.n_features,
                                  self.n_features))
                for c in range(self._n_states):
                    obsmean = np.outer(stats['observation'][c], self._means[c])

                    cvnum[c] = (means_weight * np.outer(meandiff[c],
                                                        meandiff[c])
                                + stats['observation*observation.T'][c]
                                - obsmean - obsmean.T
                                + np.outer(self._means[c], self._means[c])
                                * stats['post'][c])
                cvweight = max(covars_weight - self.n_features, 0)
                if self._cvtype == 'tied':
                    self._covars = ((covars_prior + cvnum.sum(axis=0))
                                    / (cvweight + stats['post'].sum()))
                elif self._cvtype == 'full':
                    self._covars = ((covars_prior + cvnum)
                                   / (cvweight + stats['post'][:, None, None]))


class MultinomialHMM(_BaseHMM):
    """Hidden Markov Model with multinomial (discrete) emissions

    Attributes
    ----------
    n_states : int (read-only)
        Number of states in the model.
    n_symbols : int
        Number of possible symbols emitted by the model (in the observations).
    transmat : array, shape (`n_states`, `n_states`)
        Matrix of transition probabilities between states.
    startprob : array, shape ('n_states`,)
        Initial state occupation distribution.
    emissionprob: array, shape ('n_states`, 'n_symbols`)
        Probability of emitting a given symbol when in each state.

    Methods
    -------
    eval(X)
        Compute the log likelihood of `X` under the HMM.
    decode(X)
        Find most likely state sequence for each point in `X` using the
        Viterbi algorithm.
    rvs(n=1)
        Generate `n` samples from the HMM.
    init(X)
        Initialize HMM parameters from `X`.
    fit(X)
        Estimate HMM parameters from `X` using the Baum-Welch algorithm.
    predict(X)
        Like decode, find most likely state sequence corresponding to `X`.
    score(X)
        Compute the log likelihood of `X` under the model.

    Examples
    --------
    >>> MultinomialHMM(n_states=2)
    ... #doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    MultinomialHMM(transmat=array([[ 0.5,  0.5],
           [ 0.5,  0.5]]),
            startprob_prior=1.0, n_states=2, startprob=array([ 0.5,  0.5]),
           transmat_prior=1.0)

    See Also
    --------
    GaussianHMM : HMM with Gaussian emissions
    """

    def __init__(self, n_states=1, startprob=None, transmat=None,
                 startprob_prior=None, transmat_prior=None):
        """Create a hidden Markov model with multinomial emissions.

        Parameters
        ----------
        n_states : int
            Number of states.
        """
        super(MultinomialHMM, self).__init__(n_states, startprob, transmat,
                                             startprob_prior=startprob_prior,
                                             transmat_prior=transmat_prior)

    def _get_emissionprob(self):
        """Emission probability distribution for each state."""
        return np.exp(self._log_emissionprob)

    def _set_emissionprob(self, emissionprob):
        emissionprob = np.asanyarray(emissionprob)
        if hasattr(self, 'n_symbols') and \
               emissionprob.shape != (self._n_states, self.n_symbols):
            raise ValueError('emissionprob must have shape '
                             '(n_states, n_symbols)')

        self._log_emissionprob = np.log(emissionprob)
        underflow_idx = np.isnan(self._log_emissionprob)
        self._log_emissionprob[underflow_idx] = -np.Inf
        self.n_symbols = self._log_emissionprob.shape[1]

    emissionprob = property(_get_emissionprob, _set_emissionprob)

    def _compute_log_likelihood(self, obs):
        return self._log_emissionprob[:, obs].T

    def _generate_sample_from_state(self, state):
        cdf = np.cumsum(self.emissionprob[state, :])
        rand = np.random.rand()
        symbol = (cdf > rand).argmax()
        return symbol

    def _init(self, obs, params='ste'):
        super(MultinomialHMM, self)._init(obs, params=params)

        if 'e' in params:
            emissionprob = normalize(np.random.rand(self._n_states,
                                                    self.n_symbols), 1)
            self.emissionprob = emissionprob

    def _initialize_sufficient_statistics(self):
        stats = super(MultinomialHMM, self)._initialize_sufficient_statistics()
        stats['observation'] = np.zeros((self._n_states, self.n_symbols))
        return stats

    def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                          posteriors, fwdlattice, bwdlattice,
                                          params):
        super(MultinomialHMM, self)._accumulate_sufficient_statistics(
            stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice,
            params)
        if 'e' in params:
            for t, symbol in enumerate(obs):
                stats['observation'][:, symbol] += posteriors[t, :]

    def _do_mstep(self, stats, params, **kwargs):
        super(MultinomialHMM, self)._do_mstep(stats, params)
        if 'e' in params:
            self.emissionprob = (stats['observation']
                                 / stats['observation'].sum(1)[:, np.newaxis])


class GMMHMM(_BaseHMM):
    """Hidden Markov Model with Gaussin mixture emissions

    Attributes
    ----------
    n_states : int (read-only)
        Number of states in the model.
    transmat : array, shape (`n_states`, `n_states`)
        Matrix of transition probabilities between states.
    startprob : array, shape ('n_states`,)
        Initial state occupation distribution.
    gmms: array of GMM objects, length 'n_states`
        GMM emission distributions for each state

    Methods
    -------
    eval(X)
        Compute the log likelihood of `X` under the HMM.
    decode(X)
        Find most likely state sequence for each point in `X` using the
        Viterbi algorithm.
    rvs(n=1)
        Generate `n` samples from the HMM.
    init(X)
        Initialize HMM parameters from `X`.
    fit(X)
        Estimate HMM parameters from `X` using the Baum-Welch algorithm.
    predict(X)
        Like decode, find most likely state sequence corresponding to `X`.
    score(X)
        Compute the log likelihood of `X` under the model.

    Examples
    --------
    >>> GMMHMM(n_states=2, n_mix=10, cvtype='diag')
    ... # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    GMMHMM(n_mix=10, cvtype='diag', n_states=2, startprob_prior=1.0,
        startprob=array([ 0.5,  0.5]),
        transmat=array([[ 0.5,  0.5],
           [ 0.5,  0.5]]),
        transmat_prior=1.0,
        gmms=[GMM(cvtype='diag', n_states=10), GMM(cvtype='diag',
              n_states=10)])

    See Also
    --------
    GaussianHMM : HMM with Gaussian emissions
    """

    def __init__(self, n_states=1, n_mix=1, startprob=None,
                 transmat=None, startprob_prior=None, transmat_prior=None,
                 gmms=None, cvtype=None):
        """Create a hidden Markov model with GMM emissions.

        Parameters
        ----------
        n_states : int
            Number of states.
        """
        super(GMMHMM, self).__init__(n_states, startprob, transmat,
                                     startprob_prior=startprob_prior,
                                     transmat_prior=transmat_prior)

        # XXX: Hotfit for n_mix that is incompatible with the scikit's
        # BaseEstimator API
        self.n_mix = n_mix
        self.cvtype = cvtype
        if gmms is None:
            gmms = []
            for x in range(self.n_states):
                if cvtype is None:
                    g = GMM(n_mix)
                else:
                    g = GMM(n_mix, cvtype=cvtype)
                gmms.append(g)
        self.gmms = gmms

    def _compute_log_likelihood(self, obs):
        return np.array([g.score(obs) for g in self.gmms]).T

    def _generate_sample_from_state(self, state):
        return self.gmms[state].rvs(1).flatten()

    def _init(self, obs, params='stwmc'):
        super(GMMHMM, self)._init(obs, params=params)

        allobs = np.concatenate(obs, 0)
        for g in self.gmms:
            g.fit(allobs, n_iter=0, init_params=params)

    def _initialize_sufficient_statistics(self):
        stats = super(GMMHMM, self)._initialize_sufficient_statistics()
        stats['normalization'] = [np.zeros(g.weights.shape) for g in self.gmms]
        stats['means'] = [np.zeros(np.shape(g.means)) for g in self.gmms]
        stats['covars'] = [np.zeros(np.shape(g._covars)) for g in self.gmms]
        return stats

    def _accumulate_sufficient_statistics(self, stats, obs, framelogprob,
                                          posteriors, fwdlattice, bwdlattice,
                                          params):
        super(GMMHMM, self)._accumulate_sufficient_statistics(
            stats, obs, framelogprob, posteriors, fwdlattice, bwdlattice,
            params)

        for state, g in enumerate(self.gmms):
            gmm_logprob, gmm_posteriors = g.eval(obs)
            gmm_posteriors *= posteriors[:, state][:, np.newaxis]
            tmpgmm = GMM(g.n_states, cvtype=g.cvtype)
            tmpgmm.n_features = g.n_features
            tmpgmm.covars = _distribute_covar_matrix_to_match_cvtype(
                np.eye(g.n_features), g.cvtype, g.n_states)
            norm = tmpgmm._do_mstep(obs, gmm_posteriors, params)

            stats['normalization'][state] += norm
            if 'm' in params:
                stats['means'][state] += tmpgmm.means * norm[:, np.newaxis]
            if 'c' in params:
                if tmpgmm.cvtype == 'tied':
                    stats['covars'][state] += tmpgmm._covars * norm.sum()
                else:
                    cvnorm = np.copy(norm)
                    shape = np.ones(tmpgmm._covars.ndim)
                    shape[0] = np.shape(tmpgmm._covars)[0]
                    cvnorm.shape = shape
                    stats['covars'][state] += tmpgmm._covars * cvnorm

    def _do_mstep(self, stats, params, covars_prior=1e-2, **kwargs):
        super(GMMHMM, self)._do_mstep(stats, params)
        # All that is left to do is to apply covars_prior to the
        # parameters updated in _accumulate_sufficient_statistics.
        for state, g in enumerate(self.gmms):
            norm = stats['normalization'][state]
            if 'w' in params:
                g.weights = normalize(norm)
            if 'm' in params:
                g.means = stats['means'][state] / norm[:, np.newaxis]
            if 'c' in params:
                if g.cvtype == 'tied':
                    g.covars = ((stats['covars'][state]
                                 + covars_prior * np.eye(g.n_features))
                                / norm.sum())
                else:
                    cvnorm = np.copy(norm)
                    shape = np.ones(g._covars.ndim)
                    shape[0] = np.shape(g._covars)[0]
                    cvnorm.shape = shape
                    if g.cvtype == 'spherical' or g.cvtype == 'diag':
                        g.covars = (stats['covars'][state]
                                    + covars_prior) / cvnorm
                    elif g.cvtype == 'full':
                        eye = np.eye(g.n_features)
                        g.covars = ((stats['covars'][state]
                                     + covars_prior * eye[np.newaxis, :, :])
                                    / cvnorm)
